from pyaedt import aedt_exception_handler
from pyaedt.generic.DataHandlers import json_to_dict
from pyaedt.modeler.multiparts import MultiPartComponent, Actor


def read_actors(fn, actor_lib):
    """Read and map actors in a JSON file to a list of actor objects.

    Parameters
    ----------
    fn: str
        Name of the JSON file describing the actors.
    actor_lib: str
        Full path to the library containing the actor definitions.

    Returns
    -------
    list
        List of actor objects.
    """
    a = {}
    actor_dict = json_to_dict(fn)
    for name in actor_dict:
        a[name] = Actor(actor_dict[name], actor_lib, name)
    return a


class Generic(Actor, object):
    """Provides an instance of an actor.

    This class is derived from :class:`MultiPartComponent`.

    .. note::
       Motion is always forward in the X-direction of the person
       coordinate system.

    Parameters
    ----------
    actor_folder: str
        Full path to the directory containing the definition of the person.
        This can be changed later in the :class:`Person` class definition.
    speed: float or str
        Speed of the person in the X-direction.
    relative_cs_name: str
        Name of the relative coordinate system of the actor. The default is ``None``,
        in which case the global coordinate system is used.

    """

    def __init__(self, actor_folder, speed="0", relative_cs_name=None):
        """Generic class.


        """
        super(Generic, self).__init__(actor_folder, speed=speed, relative_cs_name=relative_cs_name)


class Person(Actor, object):
    """Provides an instance of an actor.

    This class is derived from :class:`MultiPartComponent`.

    .. note::
       Motion is always forward in the X-direction of the person coordinate system.

    Parameters
    ----------
    actor_folder: str, required
        Full path to the folder containing the definition of the
        person.  This can be changed later in the :class:`Person`
        class definition.
    speed: float or str
        Speed of the person in the X-direction.
    stride: float or str
        Stride length of the person. The default is "0". An example of
        entering a string, which includes units, is ``"0.8meters"``.
    relative_cs_name: str
        Name of the relative coordinate system of the actor. The
        default is ``None``, in which case the global coordinate
        system is used.

    """

    def __init__(self, actor_folder, speed="0", stride="0.8meters", relative_cs_name=None):
        """Initialize person actor."""

        super(Person, self).__init__(actor_folder, speed=speed, relative_cs_name=relative_cs_name)

        self._stride = stride

    @property
    def stride(self):
        """Stride in meters.

        Returns
        -------
        str
        """
        return self._stride

    @stride.setter
    def stride(self, s):
        self._stride = s  # TODO: Add validation to allow expressions.

    @aedt_exception_handler
    def _add_walking(self, app):
        # Update expressions for oscillation of limbs. At this point
        # we could parse p.name to handle motion (arm, leg, ...).
        for k, p in self.parts.items():
            if any(p.rot_axis):  # use this key to determine if there is motion of the part.
                if p.rot_axis[1]:  # Make sure pitch rotation is True
                    app[p.pitch_name] = p.pitch + "*sin(2*pi*(" + self.speed_name + \
                                           "/" + self.stride + ") " \
                                           + "*" + MultiPartComponent._t + ") + " + \
                                           "(" + p['compensation_angle'] + ")rad"

    @aedt_exception_handler
    def insert(self, app, motion=True):
        """Insert the person in AEDT.

        Parameters
        ----------
        app: :class:`pyaedt.hfss.Hfss`
            HFSS application.
        motion : bool, optional
            The default is ``True``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        app.add_info_message("Adding person: " + self.name, "Design")

        # Insert the component first, then set variables.
        self._insert(app)  # Place the person in the app.
        if motion:
            self._add_speed(app)
            self._add_walking(app)


class Bird(Actor, object):
    """Provides an instance of an actor.

    This class is derived from :class:`MultiPartComponent`.

    .. note::
       Motion is always forward in the X-direction.

    Parameters
    ----------
    bird_folder : str, required
        Full path to the directory containing the definition of the
        bird.  This can be changed later.
    speed : float or str
        Speed of the bird.
    flapping_rate : float or str
        Flapping rate.
    relative_cs_name : str
        Name of the relative coordinate system of the actor. The
        default is``None``, in which case the global coordinate system
        is used.

    """

    def __init__(self, bird_folder, speed="2.0", flapping_rate="50Hz", relative_cs_name=None):
        """Bike class.

        """

        super(Bird, self).__init__(bird_folder, speed=speed, relative_cs_name=relative_cs_name)
        self._flapping_rate = flapping_rate

    def _add_flying(self, app):
        # Update expressions for wheel motion:

        for k, p in self.parts.items():
            if any(p.rot_axis):  # use this key to verify that there is local motion.
                if p.rot_axis[2]:  # Flapping is roll
                    app[p.roll_name] = p['rotation'] + '* sin(2*pi*' + str(self._flapping_rate) + '*' \
                                        + MultiPartComponent._t + ")"

    @aedt_exception_handler
    def insert(self, app, motion=True):
        """Insert the bird in AEDT.

        Parameters
        ----------
        app: :class:`pyaedt.hfss.Hfss`
        motion: bool
            The default is ``True``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        app.add_info_message("Adding Vehicle: " + self.name, "Design")

        self._insert(app)  # Place the multipart component in the app.
        if motion:
            self._add_speed(app)
            self._add_flying(app)


class Vehicle(Actor, object):
    """Provides an instance of an actor.

    This class is derived from :class:`MultiPartComponent`.

   This class is derived from :class:`MultiPartComponent`.

    .. note::
        Motion is always forward in the X-direction.

    Properties
    ----------
    car_folder : str, required
        Full path to the folder containing the definition of the
        vehicle.  This can be changed later.
    speed : float or str
        Speed of the vehicle.
    relative_cs_name : str
        Name of the relative coordinate system of the actor. The
        default is ``None``, in which case the global coordinate
        system is used.

    """

    def __init__(self, car_folder, speed=10.0, relative_cs_name=None):
        """Vehicle class."""

        super(Vehicle, self).__init__(car_folder, speed=speed, relative_cs_name=relative_cs_name)

    @aedt_exception_handler
    def _add_driving(self, app):
        # Update expressions for wheel motion:
        for k, p in self.parts.items():
            if any(p.rot_axis):  # use this key to determine if there is motion of the wheel.
                if p.rot_axis[1]:  # Make sure pitch rotation is True
                    app[p.pitch_name] = "(" + MultiPartComponent._t + "*" + \
                                        self.speed_name + "/" + p['tire_radius'] \
                                        + "meter)*(180/pi)*1deg"

    @aedt_exception_handler
    def insert(self, app, motion=True):
        """Insert the vehicle in AEDT.

        Parameters
        ----------
        app: :class:`pyaedt.hfss.Hfss`
        motion: bool, optional
            The default is ``True``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        app.add_info_message("Adding vehicle: " + self.name, "Design")

        self._insert(app)  # Place the multipart component in the app.
        if motion:
            self._add_speed(app)
            self._add_driving(app)
        return True


class Radar(MultiPartComponent, object):
    """Manages the radar definition and placement in the HFSS design.

    Parameters
    ----------
    radar_folder: str
        Full path to the folder containing the radar file.
    name: str, optional
        Name of the radar file. The default is ``None``.
    motion: bool, optional
        The default is ``False``.

    """

    def __init__(self, radar_folder, name=None, motion=False,
                 use_relative_cs=False, offset=("0", "0", "0"), speed=0, relative_cs_name=None):

        self.aedt_antenna_names = []  # List of Antenna Names
        name = name.split('.')[0] if name else name  # remove suffix if any
        self._component_class = 'radar'
        super(Radar, self).__init__(radar_folder, name=name, use_relative_cs=use_relative_cs, motion=motion,
                                    offset=offset, relative_cs_name=relative_cs_name)
        self._speed_expression = str(speed) + 'm_per_sec'
        self.pair = []

    @property
    def units(self):
        """Multipart units.

        Returns
        -------
        str
            Multipart units.
        """
        return self._local_units

    @property
    def speed_name(self):
        """Speed variable name.

        Returns
        -------
        str
            Name of the speed variable.
        """
        return self.name + '_speed'

    @property
    def speed_expression(self):
        """Speed variable expression.

        Returns
        -------
        str
            Expression for the speed variable.
        """
        return self._speed_expression

    @speed_expression.setter
    def speed_expression(self, s):
        self._speed_expression = s

    @aedt_exception_handler
    def _add_speed(self, app):
        app.variable_manager.set_variable(variable_name=self.speed_name,
                                          expression=self.speed_expression,
                                          description="radar speed")
        # Update expressions for x and y position in app:
        app[self.offset_names[0]] = str(self.offset[0]) + '+' \
                                    + self.speed_name + ' * ' + MultiPartComponent._t \
                                    + '* cos(' + self.yaw_name + ')'

        app[self.offset_names[1]] = str(self.offset[1]) + '+' \
                                    + self.speed_name + ' * ' + MultiPartComponent._t \
                                    + '* sin(' + self.yaw_name + ')'

    @aedt_exception_handler
    def insert(self, app, motion=False):
        """Insert radar in the HFSS application instance.

        Parameters
        ----------
        app: class: `pyaedt.hfss.Hfss`
        motion: bool, optional
            The default is ``False``.

        Returns
        -------
        list
            List of antenna placed.
        """
        app.add_info_message("Adding radar module:  " + self.name)
        if self.use_global_cs or self.cs_name in app.modeler.oeditor.GetCoordinateSystems():
            app.modeler.set_working_coordinate_system(self.cs_name)
            if self.use_relative_cs:
                self._relative_cs_name = self.name + '_cs'
        self.position_in_app(app)
        tx_names = []
        rx_names = []
        for p in self.parts:
            antenna_object = self.parts[p].insert(app, units=self._local_units)
            self.aedt_components.append(antenna_object.antennaname)
            self.aedt_antenna_names.append(antenna_object.excitation_name)
            if p.startswith('tx'):
                tx_names.append(antenna_object.excitation_name)
            elif p.startswith('rx'):
                rx_names.append(antenna_object.excitation_name)

        # Define tx/rx pairs:
        self.pair = {}
        for tx in tx_names:
            rx_string = ''
            for rx in rx_names:
                rx_string += rx + ','
            self.pair[tx] = rx_string.strip(',')

        app.set_sbr_txrx_settings(self.pair)

        app.modeler.create_group(components=self.aedt_components,
                                 group_name=self.name)
        app.add_info_message("Group Created:  " + self.name)
        if motion:
            self._add_speed(app)
        return self.aedt_antenna_names



import warnings

from .Analysis import Analysis
from ..modeler.Model2D import Modeler2D
from ..modules.Mesh import Mesh
from ..generic.general_methods import aedt_exception_handler


class FieldAnalysis2D(Analysis):
    """Manages 2D field analysis setup in Maxwell2D and Q2D.

    This class is automatically initialized by an application call from one of
    the 2D tools. See the application function for parameter definitions.

    Parameters
    ----------
    application : str
        2D application that is to initialize the call.
    projectname : str, optional
        Name of the project to select or the full path to the project
        or AEDTZ archive to open. The default is ``None``, in which
        case an attempt is made to get an active project. If no
        projects are present, an empty project is created.
    designname : str, optional
        Name of the design to select. The default is ``None``, in
        which case an attempt is made to get an active design. If no
        designs are present, an empty design is created.
    solution_type : str, optional
        Solution type to apply to the design. The default is
        ``None``, in which case the default type is applied.
    setup_name : str, optional
        Name of the setup to use as the nominal. The default is
        ``None``, in which case the active setup is used or
        nothing is used.
    specified_version: str, optional
        Version of AEDT  to use. The default is ``None``, in which case
        the active version or latest installed version is used.
    NG : bool, optional
        Whether to run AEDT in the non-graphical mode. The default
        is ``False``, which launches AEDT in the graphical mode.
    new_desktop_session : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the
        machine. The default is ``True``.
    close_on_exit : bool, optional
        Whether to release  AEDT on exit. The default is ``False``.
    student_version : bool, optional
        Whether to enable the student version of AEDT. The default
        is ``False``.

    """

    def __init__(
        self,
        application,
        projectname,
        designname,
        solution_type,
        setup_name=None,
        specified_version=None,
        non_graphical=False,
        new_desktop_session=False,
        close_on_exit=False,
        student_version=False,
    ):

        Analysis.__init__(
            self,
            application,
            projectname,
            designname,
            solution_type,
            setup_name,
            specified_version,
            non_graphical,
            new_desktop_session,
            close_on_exit,
            student_version,
        )
        self._modeler = Modeler2D(self)
        self._mesh = Mesh(self)
        # self._post = PostProcessor(self)

    @property
    def modeler(self):
        """Modeler.

        Returns
        -------
        :class:`pyaedt.modeler.Model2D.Modeler2D`
        """
        return self._modeler

    @property
    def mesh(self):
        """Mesh.

        Returns
        -------
        :class:`pyaedt.modules.Mesh.Mesh`
        """
        return self._mesh

    # @property
    # def post(self):
    #     return self._post

    @aedt_exception_handler
    def assignmaterial(self, obj, mat):
        """Assign a material to one or more objects.

        .. deprecated:: 0.3.1
           Use :func:`FieldAnalysis2D.assign_material` instead.

        """
        # raise a DeprecationWarning.  User won't have to change anything
        warnings.warn("assignmaterial is deprecated. Use assign_material instead.", DeprecationWarning)
        self.assign_material(obj, mat)

    @aedt_exception_handler
    def assign_material(self, obj, mat):
        """Assign a material to one or more objects.

        Parameters
        ----------
        obj : str or list
            One or more objects to assign materials to.
        mat : str
            Material to assign. If this material is not present, it will be
            created.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        mat = mat.lower()
        selections = self.modeler.convert_to_selections(obj)
        arg1 = ["NAME:Selections"]
        arg1.append("Selections:="), arg1.append(selections)
        arg2 = ["NAME:Attributes"]
        arg2.append("MaterialValue:="), arg2.append(chr(34) + mat + chr(34))
        if mat in self.materials.material_keys:
            Mat = self.materials.material_keys[mat]
            Mat.update()
            if Mat.is_dielectric():
                arg2.append("SolveInside:="), arg2.append(True)
            else:
                arg2.append("SolveInside:="), arg2.append(False)
            self.modeler.oeditor.AssignMaterial(arg1, arg2)
            self._messenger.add_info_message("Assign Material " + mat + " to object " + selections)
            if type(obj) is list:
                for el in obj:
                    self.modeler.primitives[el].material_name = mat
            else:
                self.modeler.primitives[obj].material_name = mat
            return True
        elif self.materials.checkifmaterialexists(mat):
            self.materials._aedmattolibrary(mat)
            Mat = self.materials.material_keys[mat]
            if Mat.is_dielectric():
                arg2.append("SolveInside:="), arg2.append(True)
            else:
                arg2.append("SolveInside:="), arg2.append(False)
            self.modeler.oeditor.AssignMaterial(arg1, arg2)
            self._messenger.add_info_message("Assign Material " + mat + " to object " + selections)
            if type(obj) is list:
                for el in obj:
                    self.modeler.primitives[el].material_name = mat
            else:
                self.modeler.primitives[obj].material_name = mat

            return True
        else:
            self._messenger.add_error_message("Material Does Not Exists")
            return False



import os

from pyaedt import aedt_exception_handler
from pyaedt.modeler.GeometryOperators import GeometryOperators


class Part(object):
    """Helps manage 3D component placement and definition.

    Parameters
    ----------
    part_folder : str
        Path to the folder with the A3DCOMP files.
    part_dict : dict
        Defines relevant properties of the class with the following keywords:
        * 'comp_name': str, Name of the A3DCOMP file.
        * 'offset': list or str, Offset coordinate system definition relative to the parent.
        * 'rotation_cs': list or str, Rotation coordinate system relative to the parent.
        * 'rotation': str or numeric, Rotation angle.
        * 'compensation_angle': str or numeric, Initial angle.
        * 'rotation_axis': str, Rotation axis (``"X"``, ``"Y"``, or ``"Z"``).
        * 'duplicate_number': str or int, Number of instances for linear duplication.
        * 'duplicate_vector': list, Vector for duplication relative to the parent coordinate system.
     parent :  str
         The default is ``None``.
     name : str, optional
            Name of the A3DCOMP file without the extension. The default is ``None``.
     """

    # List of known keys for a Part and default values:
    allowed_keys = {'comp_name': None,  # *.a3dcomp file name
                     'offset': None,
                     'rotation_cs': None,
                     'rotation': 0.0,
                     'compensation_angle': None,
                     'rotation_axis': 'Z',
                     'tire_radius': None,
                     'duplicate_number': None,
                     'duplicate_vector': None,
                     'antenna_type': None,  # Antenna only
                     'ffd_name': None,  # Antenna only
                     'mode': None,  # Antenna only
                     'aedt_name': None,
                     'beamwidth_elevation': None,  # Antenna only
                     'beamwidth_azimuth': None,  # Antenna only
                     'polarization': None}  # Antenna only

    def __init__(self, part_folder, part_dict, parent=None, name=None):

        # Default values:
        self._compdef = dict()
        self._parent = parent

        # Extract the 3d component name and part folder
        # from the file name.
        # Use this as the default value for comp_name.  Ensure that the correct extension is used.
        self._compdef['part_folder'] = part_folder
        for k in Part.allowed_keys:
            if k in part_dict:
                self._compdef[k] = part_dict[k]
            else:
                self._compdef[k] = Part.allowed_keys[k]

        self._motion = False
        if parent:  # Inherit _motion directly from parent.
            self._motion = self._parent.motion

        # make sure self._name is unique if it is not passed as an argument.
        if name:
            self._name = name  # Part name should be unique. No checking here.
        elif 'name' in part_dict:
            self._name = part_dict['name']
        else:
            self._name = 'radar'  # TODO: Need to fix this!

        # Update self._compdef from the library definition in the *.json file.

        for kw, val in part_dict.items():
            if kw in self._compdef:
                self._compdef[kw] = val
            else:
                raise KeyError('Key ' + kw + ' not allowed.')

        # Instantiate yaw, pitch and roll.  Might want to change
        # how this is handled. Make "rotation" a list instead of
        # using .yaw, .pitch, .roll properties?
        self.rot_axis = [False, False, False]  # [X, Y, Z] rotation Boolean
        if self._compdef['rotation_axis']:
            rotations_axis = self._compdef['rotation_axis'].split(',')
            if self._compdef['rotation']:
                rotations = self._compdef['rotation'].split(',')
            else:
                rotations = []
            y = "0"
            p = "0"
            r = "0"
            for a in rotations:
                if rotations_axis[rotations.index(a)].lower() == 'x':  # roll
                    r = a
                    self.rot_axis[2] = True
                elif rotations_axis[rotations.index(a)].lower() == 'y':  # pitch
                    p = a
                    self.rot_axis[1] = True
                elif rotations_axis[rotations.index(a)].lower() == 'z':  # yaw
                    y = a
                    self.rot_axis[0] = True

            self._yaw = y
            self._pitch = p
            self._roll = r
        else:
            self._yaw = "0"
            self._pitch = "0"
            self._roll = "0"

    def __setitem__(self, key, value):
        self._compdef[key] = value

    def __getitem__(self, key):
        if key == 'rotation_cs':
            cs = self._compdef[key]
            if cs == "Global" or cs is None:
                self._compdef[key] = ["0", "0", "0"]
            else:
                self._compdef[key] = [str(i) if not i is str else i for i in cs]
        return self._compdef[key]

    @aedt_exception_handler
    def zero_offset(self, kw):  # Returns True if cs at kw is at [0, 0, 0]
        """Return zero if the coordinate system defined by kw is [0, 0, 0].

        Parameters
        ----------
        kw : str
             'offset' or 'rotation_cs'

        Returns
        -------
        ``True`` when the coordinate system is ``[0, 0, 0]``, ``None`` otherwise.

        """
        if kw in ['offset', 'rotation_cs']:
            s = []
            if self[kw]:
                s = [GeometryOperators.is_small(c) for c in self[kw]]
            if len(s) > 0:
                return all(s)
            else:
                return True
        return False

    @property
    def file_name(self):
        """Antenna file name.

        Returns
        -------
        str
            Full name of the A3DCOMP file.
        """
        return os.path.join(self._compdef['part_folder'], self['comp_name'])

    # Create a unique coordinate system name for the part.
    @property
    def cs_name(self):
        """Coordinate system name.

        Returns
        -------
        str
            Name of the coordinat system.
        """
        if self._motion or not self.zero_offset('offset') or not self.zero_offset('rotation_cs'):
            return self.name + '_cs'
        else:
            return self._parent.cs_name

    # Define the variable names for angles in the app:
    @property
    def yaw_name(self):
        """Yaw variable name. Yaw is the rotation about the object's Z-axis.

        Returns
        -------
        str
            ame of the yaw variable.

        """
        return self.name + '_yaw'

    @property
    def pitch_name(self):
        """Pitch variable name. Pitch is the rotation about the object's Y-axis.

        Returns
        -------
        str
            Name of the pitch variable.
        """
        return self.name + '_pitch'

    @property
    def roll_name(self):
        """Roll variable name. Roll is the rotation about the object's X-axis.

        Returns
        -------
        str
             Name of the roll variable.
        """
        return self.name + '_roll'

    # Always return the local origin as a list:
    @property
    def local_origin(self):
        """Local part offset.

        Returns
        -------
        list
            List of offset values for the local part.
        """
        if self['offset']:
            if self.zero_offset('offset') or self['offset'] == 'Global':
                return [0, 0, 0]
            else:
                if self._parent._local_units:
                    units = self._parent._local_units
                else:
                    units = self._parent.modeler_units
                offset = [str(i)+units for i in self['offset']]

                return offset
        else:
            return [0, 0, 0]

    @property
    def rotate_origin(self):
        """Origin rotation list.

        Returns
        -------
        list
            List of offset values for the rotation.
        """
        if self['rotation_cs']:
            if self.zero_offset('rotation_cs') or self['rotation_cs'] == 'Global':
                return self.local_origin
            else:
                return self['rotation_cs']
        else:
            return [0, 0, 0]

    @property
    def _do_rotate(self):  # True if any rotation angles are non-zero or 'rotation_cs' is defined.
        return any(self.rot_axis)

    @property
    def _do_offset(self):  # True if any rotation angles are non-zero.
        return any(GeometryOperators.numeric_cs(self.local_origin))

    # Allow expressions should be valid angle as either string
    # or numerical value.
    @property
    def yaw(self):
        """Yaw variable value.

        Returns
        -------
        str
            Value for the yaw variable.
        """
        return self._yaw

    @yaw.setter
    def yaw(self, yaw):
        self._yaw = yaw

    @property
    def pitch(self):
        """Pitch variable value.

        Returns
        -------
        str
            Value of the pitch variable.
        """
        return self._pitch

    @pitch.setter
    def pitch(self, pitch):
        self._pitch = pitch

    @property
    def roll(self):
        """Roll variable value.

        Returns
        -------
        str
            Value of the roll variable.
        """
        return self._roll

    @roll.setter
    def roll(self, roll):
        self._roll = roll

    @property
    def name(self):
        """Part name.

        Returns
        -------
        str
            Name of the part.
        """
        return self._parent.name + '_' + self._name

    @aedt_exception_handler
    def set_relative_cs(self, app):
        """Create a new, parametric Coordinate System.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.
        """
        # Set x,y,z offset variables in app. But check first to see if the CS
        # has already been defined.
        if self.cs_name not in app.modeler.oeditor.GetCoordinateSystems() and self.cs_name != "Global":
            x_pointing = [1, 0, 0]
            y_pointing = [0, 1, 0]
            app.modeler.create_coordinate_system(origin=self.local_origin,
                                                 x_pointing=x_pointing,
                                                 y_pointing=y_pointing,
                                                 reference_cs=self._parent.cs_name,
                                                 mode="axis",
                                                 name=self.cs_name)
        return True

    @property
    def rot_cs_name(self):
        """Rotation coordinate system name.

        Returns
        -------
        str
            Name of the rotation coordinate system.
        """
        return self.name + '_rot_cs'

    @aedt_exception_handler
    def do_rotate(self, app, aedt_object):
        """Set the rotation coordinate system relative to the parent coordinate system.

        This method should only be called if there is rotation in the component.
        The rotation coordinate system is offset from the parent coordinate system.

        Parameters
        ----------
        app: pyaedt.Hfss
            HFSS instance of AEDT.
        aedt_object : str
            Name of the design in AEDT.
        """

        x_pointing = [1, 0, 0]
        y_pointing = [0, 1, 0]
        app.modeler.create_coordinate_system(origin=self.rotate_origin,
                                             x_pointing=x_pointing,
                                             y_pointing=y_pointing,
                                             reference_cs=self._parent.cs_name,
                                             mode="axis",
                                             name=self.rot_cs_name)
        if self.rot_axis[0]:
            app[self.yaw_name] = self.yaw
            app.modeler.rotate(aedt_object, 'Z', angle=self.yaw_name)
        if self.rot_axis[1]:
            app[self.pitch_name] = self.pitch
            app.modeler.rotate(aedt_object, 'Y', angle=self.pitch_name)
        if self.rot_axis[2]:
            app[self.roll_name] = self.roll
            app.modeler.rotate(aedt_object, 'X', angle=self.roll_name)

        return True

    @aedt_exception_handler
    def insert(self, app):
        """Insert 3D component in AEDT.

        Parameters
        ----------
        app: class:`pyaedt.hfss.Hfss`
            HFSS application instance.

        Returns
        -------
        str
            Name of inserted object.
        """
        aedt_objects = []
        # TODO: Why the inconsistent syntax for cs commands?
        if self._do_offset:
            self.set_relative_cs(app)  # Create coordinate system, if needed.
            aedt_objects.append(app.modeler.primitives.insert_3d_component(self.file_name, targetCS=self.cs_name))
        else:
            aedt_objects.append(
                app.modeler.primitives.insert_3d_component(self.file_name, targetCS=self._parent.cs_name))
        if self._do_rotate:
            self.do_rotate(app, aedt_objects[0])

        # Duplication occurs in parent coordinate system.
        app.modeler.set_working_coordinate_system(self._parent.cs_name)
        if self['duplicate_vector']:
            d_vect = [float(i) for i in self['duplicate_vector']]
            duplicate_result = app.modeler.duplicate_along_line(aedt_objects[0], d_vect,
                                                                nclones=int(self['duplicate_number']), is_3d_comp=True)
            if duplicate_result[0]:
                for d in duplicate_result[1]:
                    aedt_objects.append(d)
        return aedt_objects


class Antenna(Part, object):
    """Manages antennas.

    This class is derived from :class:`Part`.

    Parameters
    ----------
    root_folder:

    ant_dict:

    parent: str, optional
        The default is ``None``.
    name: str, optional
        The default is ``None``.

    """
    def __init__(self, root_folder, ant_dict, parent=None, name=None):
        super(Antenna, self).__init__(root_folder, ant_dict, parent=parent, name=name)

    def _antenna_type(self, app):
        if self._compdef['antenna_type'] == 'parametric':
            return app.SbrAntennas.ParametricBeam
        if self._compdef['antenna_type'] == 'ffd':
            return 'file'

    @property
    def params(self):
        """Multi-part parameters.

        Returns
        -------
        dict
            Dictionary of parameters for a multi-part.
        """
        p = {}
        if self._compdef['antenna_type'] == 'parametric':
            p['Vertical BeamWidth'] = self._compdef['beamwidth_elevation']
            p['Horizontal BeamWidth'] = self._compdef['beamwidth_azimuth']
            p['Polarization'] = self._compdef['polarization']
        return p

    @aedt_exception_handler
    def _insert(self, app, target_cs=None, units=None):
        if not target_cs:
            target_cs = self._parent.cs_name
        if not units:
            if self._parent._local_units:
                units = self._parent._local_units
            else:
                units = self._parent.units
        if self._compdef['ffd_name']:
            ffd = os.path.join(self._compdef['part_folder'], self._parent._name + ".ffd", self._compdef['ffd_name'])
            a = app.create_sbr_file_based_antenna(ffd_full_path=ffd, model_units=units,
                                       target_cs=target_cs,
                                       antenna_name=self.name)
        else:
            a = app.create_sbr_antenna(self._antenna_type(app),
                                       model_units=units,
                                       parameters_dict=self.params,
                                       target_cs=target_cs,
                                       antenna_name=self.name)
        return a

    @aedt_exception_handler
    def insert(self, app, units=None):
        """Insert antenna in the app.

        Parameters
        ----------
        app: :class:`pyaedt.hfss.Hfss`, required HFSS application instance.

        Returns
        -------
        str
            Name of the inserted object.
        """

        if self._do_offset:
            self.set_relative_cs(app)
            antenna_object = self._insert(app, units=units)  # Create coordinate system, if needed.
        else:
            antenna_object = self._insert(app, target_cs=self._parent.cs_name, units=units)
        if self._do_rotate and antenna_object:
            self.do_rotate(app, antenna_object.antennaname)

        return antenna_object



class BLNDChamferRangeLaw(object):
    """BLNDChamferRangeLaw class."""

    (BLNDConstantRange, BLNDVariableRange) = range(0, 2)


class BLNDFilletRadiusLaw(object):
    """BLNDFilletRadiusLaw class."""

    (BLNDConstantRadius, BLNDVariableRadius) = range(0, 2)


class BLNDFilletType(object):
    """BLNDFilletType class."""

    (BLNDRound, BLNDMitered) = range(0, 2)


class CoordinateSystemAxis(object):
    """CoordinateSystemAxis class."""

    (XAxis, YAxis, ZAxis) = range(0, 3)


class CoordinateSystemPlane(object):
    """CoordinateSystemPlane class."""

    (XYPlane, YZPlane, ZXPlane) = range(0, 3)


class Plane(object):
    """Plane class."""

    (XYPlane, YZPlane, ZXPlane, ISO) = ("XY", "YZ", "ZX", "iso")


class GravityDirection(object):
    """GravityDirection class."""

    (XNeg, YNeg, ZNeg, XPos, YPos, ZPos) = range(0, 6)


class MessageSeverity(object):
    """MessageSeverity class."""

    (InfoMessage, WarningMessage, ErrorMessage, FatalMessage, IncompleteMessage) = range(0, 5)


class ParamDataType(object):
    """ParamDataType class."""

    (Unknown, Int, Double, String, Bool, Position) = range(-1, 5)


class ParamPropFlag(object):
    """ParamPropFlag class."""

    (Unknown, NoFlag, ReadOnly, MustBeInt) = range(-1, 3)
    MustBeReal = 4
    Hidden = 8


class ParamPropType(object):
    """ParamPropTyp class."""

    (Unknown, Text, Menu, Number, Value, FileName, Checkbox, Position) = range(-1, 7)


class PartPropertyFlags(object):
    """PartPropertyFlags class."""

    PropNonModel = 1
    PropDisplayWireFrame = 2
    PropReadOnly = 4
    PostprocessingGeometry = 8
    PropInvisible = 16
    PropShowDirection = 32
    PropDummy = 33


class PolylineSegmentType(object):
    """PolylineSegmentType class."""

    (LineSegment, ArcSegment, SplineSegment, AngularArcSegment) = range(0, 4)


class PolylineXSegmentType(object):
    """PolylineXSegmentType class."""

    (NONE, Line, Circle, Rectangle, Trapezoid) = range(0, 5)


class SplitWhichSideToKeep(object):
    """SplitWhichSideToKeep class."""

    (SplitKeepBoth, SplitKeepPositiveOnly, SplitKeepNegativeOnly) = range(0, 3)


class SweepDraftType(object):
    """SweepDraftType class."""

    (ExtendedDraft, RoundDraft, NaturalDraft, MixedDraft) = range(0, 4)


class UnitType(object):
    """UnitType  class."""

    (NoUnit, LengthUnit, AngleUnit) = range(0, 3)



"""
This module contains the `Desktop` class.

This module is used to initialize AEDT and Message Manager to manage AEDT.

You can initialize this module before launching an app or
have the app automatically initialize it to the latest installed AEDT version.
"""
from __future__ import absolute_import

import os
import sys
import traceback
import logging
import pkgutil
import getpass
import re
import warnings
import gc
import time
import datetime
import tempfile
from pyaedt.application.MessageManager import AEDTMessageManager
from pyaedt.misc import list_installed_ansysem
from pyaedt import is_ironpython, _pythonver, inside_desktop


pathname = os.path.dirname(__file__)
if os.path.exists(os.path.join(pathname, "version.txt")):
    with open(os.path.join(pathname, "version.txt"), "r") as f:
        pyaedtversion = f.readline().strip()
elif os.path.exists(os.path.join(pathname, "..", "version.txt")):
    with open(os.path.join(pathname, "..", "version.txt"), "r") as f:
        pyaedtversion = f.readline().strip()
else:
    pyaedtversion = "X"


if os.name == "nt":
    IsWindows = True
else:
    IsWindows = False
logger = logging.getLogger(__name__)

if is_ironpython:
    import clr  # IronPython C:\Program Files\AnsysEM\AnsysEM19.4\Win64\common\IronPython\ipy64.exe

    _com = "ironpython"
elif IsWindows:
    import pythoncom

    modules = [tup[1] for tup in pkgutil.iter_modules()]
    if "clr" in modules:
        import clr
        import win32com.client

        _com = "pythonnet_v3"
    elif "win32com" in modules:
        import win32com.client

        _com = "pywin32"
    else:
        raise Exception("Error. No win32com.client or Pythonnet modules found. Please install them")


def exception_to_desktop(self, ex_value, tb_data):
    """Writes the trace stack to the desktop when a Python error occurs.

    The message is added to the AEDT global Message Manager and to the log file (if present).

    Parameters
    ----------
    ex_value : str
        Type of exception.
    tb_data : str
        Traceback information.

    """
    desktop = sys.modules["__main__"].oDesktop
    try:
        oproject = desktop.GetActiveProject()
        project_name = oproject.GetName()
        try:
            design_name = oproject.GetActiveDesign().GetName()
            if ";" in design_name:
                design_name = design_name.split(";")[1]
        except:
            design_name = ""
    except:
        project_name = ""
        design_name = ""
    tb_trace = traceback.format_tb(tb_data)
    tblist = tb_trace[0].split("\n")
    desktop.AddMessage(project_name, design_name, 2, str(ex_value))
    for el in tblist:
        desktop.AddMessage(project_name, design_name, 2, el)


def update_aedt_registry(key, value, desktop_version="211"):
    """Update the AEDT registry key.

    .. note::
       This method is only supported on Windows.

    Parameters
    ----------
    key : str
        Registry key.
    value : str
        Value for the registry key. The value includes "" if needed.
    desktop_version : str, optional
        Version of AEDT to use. The default is ``"211"``
        to use 2021 R1.

    Examples
    --------
    Update the HPC license type for HFSS in the AEDT registry.

    >>> update_aedt_registry("HFSS/HPCLicenseType", "12") # doctest: +SKIP

    Update the HPC license type for Icepak in the AEDT registry.

    >>> update_aedt_registry("Icepak/HPCLicenseType", "8") # doctest: +SKIP

    Update the legacy HPC license type for HFSS in the AEDT registry.

    >>> update_aedt_registry("HFSS/UseLegacyElectronicsHPC", "0") # doctest: +SKIP

    Update the MPI vendor for HFSS in the AEDT registry.

    >>> update_aedt_registry("HFSS/MPIVendor", "Intel") # doctest: +SKIP

    """
    if os.name == "posix":
        import subprocessdotnet as subprocess
    else:
        import subprocess
    desktop_install_dir = os.environ["ANSYSEM_ROOT" + str(desktop_version)]

    with open(os.path.join(desktop_install_dir, "config", "ProductList.txt")) as file:
        product_version = next(file).rstrip()  # get first line

    options = '-set -ProductName {} + product_version -RegistryKey "{}" -RegistryValue "{}"'.format(
        product_version, key, value
    )
    command = '"{}/UpdateRegistry" {}'.format(desktop_install_dir, options)

    subprocess.call([command])


def _delete_objects():
    module = sys.modules["__main__"]
    if "COMUtil" in dir(module):
        del module.COMUtil
    if "Hfss" in dir(module):
        del module.Hfss
    if "Edb" in dir(module):
        del module.Edb
    if "Q3d" in dir(module):
        del module.Q3d
    if "Q2d" in dir(module):
        del module.Q2d
    if "Maxwell3d" in dir(module):
        del module.Maxwell3d
    if "Maxwell2d" in dir(module):
        del module.Maxwell2d
    if "Icepak" in dir(module):
        del module.Icepak
    if "Mechanical" in dir(module):
        del module.Mechanical
    if "Emit" in dir(module):
        del module.Emit
    if "Circuit" in dir(module):
        del module.Circuit
    if "Simplorer" in dir(module):
        del module.Simplorer
    if "Hfss3dLayout" in dir(module):
        del module.Hfss3dLayout
    if "oMessenger" in dir(module):
        del module.oMessenger
    if "oDesktop" in dir(module):
        del module.oDesktop
    if "pyaedt_initialized" in dir(module):
        del module.pyaedt_initialized
    gc.collect()


def release_desktop(close_projects=True, close_desktop=True):
    """Release the AEDT API.

    Parameters
    ----------
    close_projects : bool, optional
        Whether to close the AEDT projects open in the session. The default is ``True``.
    close_desktop : bool, optional
        Whether to close the active AEDT session. The default is ``True``.

    Returns
    -------
    bool
        ``True`` when successful, ``False`` when failed.

    """

    Module = sys.modules["__main__"]
    if "oDesktop" not in dir(Module):
        _delete_objects()
        return False
    else:
        desktop = Module.oDesktop
        if close_projects:
            projects = desktop.GetProjectList()
            for project in projects:
                desktop.CloseProject(project)
        pid = Module.oDesktop.GetProcessID()
        if not (is_ironpython and inside_desktop):
            i = 0
            scopeID = 5
            while i <= scopeID:
                Module.COMUtil.ReleaseCOMObjectScope(Module.COMUtil.PInvokeProxyAPI, i)
                i += 1
            _delete_objects()

        if close_desktop:
            try:
                os.kill(pid, 9)
                _delete_objects()
                return True
            except:
                warnings.warn("Something went wrong in Closing AEDT")
                return False
    return True


def force_close_desktop():
    """Forcibly close all AEDT projects and shut down AEDT.

    Returns
    -------
    bool
        ``True`` when successful, ``False`` when failed.

    """
    Module = sys.modules["__main__"]
    pid = Module.oDesktop.GetProcessID()
    if pid > 0:
        try:
            projects = Module.oDesktop.GetProjectList()
            for project in projects:
                Module.oDesktop.CloseProject(project)
        except:
            logger.warning("No Projects. Closing Desktop Connection")
        try:
            i = 0
            scopeID = 5
            while i <= scopeID:
                Module.COMUtil.ReleaseCOMObjectScope(Module.COMUtil.PInvokeProxyAPI, 0)
                i += 1
        except:
            logger.warning("No COM UTIL. Closing the Desktop....")
        try:
            del Module.pyaedt_initialized
        except:
            pass
        try:
            os.kill(pid, 9)
            del Module.oDesktop
            successfully_closed = True
        except:
            Module.oMessenger.add_error_message("Something went wrong in Closing AEDT.")
            successfully_closed = False
        finally:
            log = logging.getLogger(__name__)
            handlers = log.handlers[:]
            for handler in handlers:
                handler.close()
                log.removeHandler(handler)
            return successfully_closed


class Desktop:
    """Initializes AEDT based on the inputs provided.

    .. note::
       On Windows, this class works without limitations in IronPython and CPython.
       On Linux, this class works only in embedded IronPython in AEDT.

    Parameters
    ----------
    specified_version : str, optional
        Version of AEDT to use. The default is ``None``, in which case the
        active setup or latest installed version is used.
    non_graphical: bool, optional
        Whether to launch AEDT in the non-graphical mode. The default
        is ``False``, in which case AEDT is launched in the graphical mode.
    new_desktop_session : bool, optional
        Whether to launch an instance of AEDT in a new thread, even if
        another instance of the ``specified_version`` is active on the machine.
        The default is ``True``.
    close_on_exit : bool, optional
        Whether to close AEDT on exit. The default is ``True``.
    student_version : bool, optional
        Whether to open the AEDT student version. The default is
        ``False``.

    Examples
    --------
    Launch AEDT 2021 R1 in non-graphical mode and initialize HFSS.

    >>> import pyaedt
    >>> desktop = pyaedt.Desktop("2021.1", non_graphical=True)
    pyaedt Info: pyaedt v...
    pyaedt Info: Python version ...
    >>> hfss = pyaedt.Hfss(designname="HFSSDesign1")
    pyaedt Info: Project...
    pyaedt Info: Added design 'HFSSDesign1' of type HFSS.

    Launch AEDT 2021 R1 in graphical mode and initialize HFSS.

    >>> desktop = Desktop("2021.1")
    pyaedt Info: pyaedt v...
    pyaedt Info: Python version ...
    >>> hfss = pyaedt.Hfss(designname="HFSSDesign1")
    pyaedt Info: No project is defined. Project...
    """

    def __init__(self, specified_version=None, non_graphical=False, new_desktop_session=True, close_on_exit=True,
                 student_version=False):
        """Initialize desktop."""
        self._main = sys.modules["__main__"]
        self._main.interpreter = _com
        self.close_on_exit = close_on_exit
        self._main.isoutsideDesktop = False
        self._main.pyaedt_version = pyaedtversion
        self._main.interpreter_ver = _pythonver
        self.releae_on_exit = True
        self.logfile = None
        if "oDesktop" in dir(self._main) and self._main.oDesktop is not None:
            self.releae_on_exit = False
        else:
            if "oDesktop" in dir(self._main):
                del self._main.oDesktop
            version_student, version_key, version = self._set_version(specified_version, student_version)
            if _com == 'ironpython':
                print("Launching PyAEDT outside Electronics Desktop with IronPython")
                self._init_ironpython(non_graphical, new_desktop_session, version)
            elif _com == 'pythonnet_v3':
                print("Launching PyAEDT outside Electronics Desktop with CPython and Pythonnet")
                self._init_cpython(non_graphical, new_desktop_session, version, student_version, version_key)
            else:
                self.add_info_message("Launching PyAEDT outside AEDT with CPython and PyWin32.")
                oAnsoftApp = win32com.client.Dispatch(version)
                self._main.oDesktop = oAnsoftApp.GetAppDesktop()
                self._main.isoutsideDesktop = True
            self._main.AEDTVersion = version_key
        self._init_logger()
        self._init_desktop()
        self._main.oMessenger.add_info_message("pyaedt v{}".format(self._main.pyaedt_version))
        self._main.oMessenger.add_info_message("Python version {}".format(sys.version))

    def __enter__(self):
        return self

    def __exit__(self, ex_type, ex_value, ex_traceback):
        # Write the trace stack to the log file if an exception occurred in the main script.
        if ex_type:
            err = self._exception(ex_value, ex_traceback)
        if self.releae_on_exit:
            self.release_desktop(close_projects=self.close_on_exit, close_on_exit=self.close_on_exit)

    @property
    def install_path(self):
        """Installation path for AEDT."""
        version_key = self._main.AEDTVersion
        root = self._version_ids[version_key]
        return os.environ[root]

    @property
    def version_keys(self):
        """Version keys for AEDT."""

        self._version_keys = []
        self._version_ids = {}
        version_list = list_installed_ansysem()
        for version_env_var in version_list:
            if "ANSYSEMSV_ROOT" in version_env_var:
                current_version_id = version_env_var.replace("ANSYSEMSV_ROOT", "")
                student = True
            else:
                current_version_id = version_env_var.replace("ANSYSEM_ROOT", "")
                student = False
            version = int(current_version_id[0:2])
            release = int(current_version_id[2])
            if version < 20:
                if release < 3:
                    version -= 1
                else:
                    release -= 2
            if student:
                v_key = "20{0}.{1}SV".format(version, release)
                self._version_keys.append(v_key)
                self._version_ids[v_key] = version_env_var
            else:
                v_key = "20{0}.{1}".format(version, release)
                self._version_keys.append(v_key)
                self._version_ids[v_key] = version_env_var

        return self._version_keys

    @property
    def current_version(self):
        """Current version of AEDT."""
        return self.version_keys[0]

    @property
    def current_version_student(self):
        """Current student version of AEDT."""
        for version_key in self.version_keys:
            if "SV" in version_key:
                return version_key
        return None

    def _init_desktop(self):
        self._main.AEDTVersion = self._main.oDesktop.GetVersion()[0:6]
        self._main.oDesktop.RestoreWindow()
        self._main.oMessenger = AEDTMessageManager()
        self._main.sDesktopinstallDirectory = self._main.oDesktop.GetExeDir()
        self._main.pyaedt_initialized = True

    def _set_version(self, specified_version, student_version):
        version_student = False
        if specified_version:
            if student_version:
                specified_version += "SV"
                version_student = True
            assert specified_version in self.version_keys, "Specified version {} not known.".format(specified_version)
            version_key = specified_version
        else:
            if student_version and self.current_version_student:
                version_key = self.current_version_student
                version_student = True
            else:
                version_key = self.current_version
                version_student = False
        if student_version and version_student:
            version = "Ansoft.ElectronicsDesktop." + version_key[:-2]
        else:
            version = "Ansoft.ElectronicsDesktop." + version_key
        self._main.sDesktopinstallDirectory = os.getenv(self._version_ids[version_key])
        self._main.AEDTVersion = version_key
        return version_student, version_key, version

    def _init_ironpython(self, non_graphical, new_aedt_session, version):
        base_path = self._main.sDesktopinstallDirectory
        sys.path.append(base_path)
        sys.path.append(os.path.join(base_path, "PythonFiles", "DesktopPlugin"))
        clr.AddReference("Ansys.Ansoft.CoreCOMScripting")
        AnsoftCOMUtil = __import__("Ansys.Ansoft.CoreCOMScripting")
        self.COMUtil = AnsoftCOMUtil.Ansoft.CoreCOMScripting.Util.COMUtil
        self._main.COMUtil = self.COMUtil
        StandalonePyScriptWrapper = AnsoftCOMUtil.Ansoft.CoreCOMScripting.COM.StandalonePyScriptWrapper
        if non_graphical or new_aedt_session:
            # forcing new thread to start in non-graphical
            oAnsoftApp = StandalonePyScriptWrapper.CreateObjectNew(non_graphical)
        else:
            oAnsoftApp = StandalonePyScriptWrapper.CreateObject(version)
        if non_graphical:
            os.environ["PYAEDT_DESKTOP_LOGS"] = "False"
        self._main.oDesktop = oAnsoftApp.GetAppDesktop()
        self._main.isoutsideDesktop = True
        return True

    def _get_tasks_list_windows(self, student_version):
        processID2 = []
        username = getpass.getuser()
        if student_version:
            process = "ansysedtsv.exe"
        else:
            process = "ansysedt.exe"
        with os.popen('tasklist /FI "IMAGENAME eq {}" /v'.format(process)) as tasks_list:
            output = tasks_list.readlines()
        pattern = r"(?i)^(?:{})\s+?(\d+)\s+.+[\s|\\](?:{})\s+".format(process, username)
        for l in output:
            m = re.search(pattern, l)
            if m:
                processID2.append(m.group(1))
        return processID2

    def _run_student(self):
        import subprocess

        DETACHED_PROCESS = 0x00000008
        pid = subprocess.Popen(
            [os.path.join(self._main.sDesktopinstallDirectory, "ansysedtsv.exe")], creationflags=DETACHED_PROCESS
        ).pid
        time.sleep(5)

    def _dispatch_win32(self, version):
        o_ansoft_app = win32com.client.Dispatch(version)
        self._main.oDesktop = o_ansoft_app.GetAppDesktop()
        self._main.isoutsideDesktop = True

    def _init_cpython(self, non_graphical, new_aedt_session, version, student_version, version_key):
        base_path = self._main.sDesktopinstallDirectory
        sys.path.append(base_path)
        sys.path.append(os.path.join(base_path, "PythonFiles", "DesktopPlugin"))
        launch_msg = "Launching AEDT installation {}".format(base_path)
        print(launch_msg)
        print("===================================================================================")
        clr.AddReference("Ansys.Ansoft.CoreCOMScripting")
        AnsoftCOMUtil = __import__("Ansys.Ansoft.CoreCOMScripting")
        self.COMUtil = AnsoftCOMUtil.Ansoft.CoreCOMScripting.Util.COMUtil
        self._main.COMUtil = self.COMUtil
        StandalonePyScriptWrapper = AnsoftCOMUtil.Ansoft.CoreCOMScripting.COM.StandalonePyScriptWrapper
        print("PyAEDT Info: Launching AEDT with module Pythonnet.")
        processID = []
        if IsWindows:
            processID = self._get_tasks_list_windows(student_version)
        if student_version and not processID:
            self._run_student()
        elif non_graphical or new_aedt_session or not processID:
            # Force new object if no non-graphical instance is running or if there is not an already existing process.
            StandalonePyScriptWrapper.CreateObjectNew(non_graphical)
        else:
            StandalonePyScriptWrapper.CreateObject(version)
        if non_graphical:
            os.environ["PYAEDT_DESKTOP_LOGS"] = "False"
        processID2 = []
        if IsWindows:
            processID2 = self._get_tasks_list_windows(student_version)
        proc = [i for i in processID2 if i not in processID]
        if not proc:
            proc = processID2
        if proc == processID2 and len(processID2) > 1:
            self._dispatch_win32(version)
        elif version_key >= "2021.1":
            if student_version:
                print("PyAEDT Info:: {} Student version started with process ID {}.".format(version, proc[0]))
            else:
                print("PyAEDT Info:: {} Started with process ID {}.".format(version, proc[0]))
            context = pythoncom.CreateBindCtx(0)
            running_coms = pythoncom.GetRunningObjectTable()
            monikiers = running_coms.EnumRunning()
            for monikier in monikiers:
                m = re.search(version[10:] + r"\.\d:" + str(proc[0]), monikier.GetDisplayName(context, monikier))
                if m:
                    obj = running_coms.GetObject(monikier)
                    self._main.isoutsideDesktop = True
                    self._main.oDesktop = win32com.client.Dispatch(obj.QueryInterface(pythoncom.IID_IDispatch))
                    break
        else:
            warnings.warn(
                "PyAEDT is not supported in AEDT versions older than 2021.1. Trying to launch it with PyWin32."
            )
            self._dispatch_win32(version)

    def _init_logger(self):
        # Set up the log file in the AEDT project directory
        self.logger = logging.getLogger(__name__)
        if not self.logger.handlers:
            if "oDesktop" in dir(self._main):
                project_dir = self._main.oDesktop.GetProjectDirectory()
            else:
                project_dir = tempfile.gettempdir()
            self.logfile = os.path.join(
                project_dir, "pyaedt{}.log".format(datetime.datetime.now().strftime("%Y%m%d_%H%M%S"))
            )
            logging.basicConfig(
                filename=self.logfile,
                format="%(asctime)s:%(name)s:%(levelname)-8s:%(message)s",
                level=logging.DEBUG,
                datefmt="%Y/%m/%d %H.%M.%S",
                filemode="w",
            )
        return True

    def _exception(self, ex_value, tb_data):
        """Write the trace stack to the desktop when a Python error occurs.

        Parameters
        ----------
        ex_value : str
            Type of the exception.
        tb_data : str
            Traceback data.

        Returns
        -------
        str
            Type of the exception.

        """
        try:
            oproject = self._main.oDesktop.GetActiveProject()
            try:
                design_name = oproject.GetActiveDesign().GetName()
                if ";" in design_name:
                    design_name = design_name.split(";")[1]
            except:
                design_name = ""
        except:
            design_name = ""
        tb_trace = traceback.format_tb(tb_data)
        tblist = tb_trace[0].split("\n")
        self._main.oMessenger.add_error_message(str(ex_value), "Global")
        for el in tblist:
            self._main.oMessenger.add_error_message(el, "Global")

        return str(ex_value)

    def release_desktop(self, close_projects=True, close_on_exit=True):
        """Release AEDT.

        Parameters
        ----------
        close_projects : bool, optional
            Whether to close the AEDT projects opened in the session.
            The default is ``True``.
        close_on_exit : bool, optional
            Whether to close the active AEDT session on exiting AEDT.
            The default is ``True``.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.release_desktop(close_projects=False, close_on_exit=False) # doctest: +SKIP

        """
        release_desktop(close_projects, close_on_exit)
        props = [a for a in dir(self) if not a.startswith("__")]
        for a in props:
            self.__dict__.pop(a, None)
        gc.collect()

    def force_close_desktop(self):
        """Forcibly close all AEDT projects and shut down AEDT.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.force_close_desktop() # doctest: +SKIP

        """
        force_close_desktop()

    def close_desktop(self):
        """Close all AEDT projects and shut down AEDT.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.close_desktop() # doctest: +SKIP

        """
        force_close_desktop()

    def enable_autosave(self):
        """Enable the auto save option.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.enable_autosave()

        """
        self._main.oDesktop.EnableAutoSave(True)

    def disable_autosave(self):
        """Disable the auto save option.

        Examples
        --------
        >>> import pyaedt
        >>> desktop = pyaedt.Desktop("2021.1")
        pyaedt Info: pyaedt v...
        pyaedt Info: Python version ...
        >>> desktop.disable_autosave()

        """
        self._main.oDesktop.EnableAutoSave(False)


def get_version_env_variable(version_id):
    """Retrieve the environment variable for the AEDT version.

    Parameters
    ----------
    version_id : str
        Full AEDT version number, such as ``"2021.1"``.

    Returns
    -------
    str
        Environment variable for the version.

    Examples
    --------
    >>> from pyaedt import desktop
    >>> desktop.get_version_env_variable("2021.1")
    'ANSYSEM_ROOT211'

    """
    version_env_var = "ANSYSEM_ROOT"
    values = version_id.split(".")
    version = int(values[0][2:])
    release = int(values[1])
    if version < 20:
        if release < 3:
            version += 1
        else:
            release += 2
    version_env_var += str(version) + str(release)
    return version_env_var



import math

from ..generic.general_methods import aedt_exception_handler
from .Modeler import Modeler, GeometryModeler
from .Primitives2D import Primitives2D


class ModelerRMxprt(Modeler):
    """Provides the Modeler RMxprt application interface.

    This class is inherited in the caller application and is accessible through the modeler variable
    object( eg. ``rmxprt.modeler``).

    """

    def __init__(self, parent):
        Modeler.__init__(self, parent)

    @property
    def oeditor(self):
        """Editor."""
        return self.odesign.SetActiveEditor("Machine")


class Modeler2D(GeometryModeler):
    """Provides the Modeler 2D application interface.

    This class is inherited in the caller application and is accessible through the modeler variable
    object( eg. ``maxwell2d.modeler``).

    Parameters
    ----------
    application :

    is3d : bool, optional
        Whether the model is 3D. The default is ``False``.

    """

    def __init__(self, application):
        GeometryModeler.__init__(self, application, is3d=False)
        self._primitives = Primitives2D(self._parent, self)
        self._primitivesDes = self._parent.project_name + self._parent.design_name

    def __get__(self, instance, owner):
        self._parent = instance
        return self

    @property
    def primitives(self):
        """Primitives.

        Returns
        -------
        :class:`pyaedt.modeler.Primitives2D.Primitives2D`

        """
        if self._primitivesDes != self._parent.project_name + self._parent.design_name:
            self._primitives.refresh()
            self._primitivesDes = self._parent.project_name + self._parent.design_name
        return self._primitives

    @aedt_exception_handler
    def calculate_radius_2D(self, object_name, inner=False):
        """Calculate the extremity of an object in the radial direction.

        Parameters
        ----------
        object_name : str
            name of the object from which to calculate the radius.
        inner : bool, optional
            The default is ``False``.

        Returns
        -------
        float
            Radius value.

            .. note::
                If ``inner=True``, then the maximum is returned; otherwise,
                the minimum is returned.

        """
        oVertexIDs = self.oeditor.GetVertexIDsFromObject(object_name)
        vert_id_int = [int(x) for x in oVertexIDs]
        if inner:
            radius = 0
        else:
            radius = 1e9

        for vertex in vert_id_int:
            pos = self.oeditor.GetVertexPosition(vertex)
            vertex_radius = math.sqrt(float(pos[0]) ** 2 + float(pos[1]) ** 2)
            if inner:
                if vertex_radius > radius:
                    radius = vertex_radius
            else:
                if vertex_radius < radius:
                    radius = vertex_radius

        return radius

    @aedt_exception_handler
    def radial_split_2D(self, radius, name):
        """Split the stator and rotor for mesh refinement.

        Parameters
        ----------
        radius : float
            Radius of the circle.
        name : str
            Name of the circle.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        self.oeditor.CreateCircle(
            [
                "NAME:CircleParameters",
                "IsCovered:=",
                True,
                "XCenter:=",
                "0mm",
                "YCenter:=",
                "0mm",
                "ZCenter:=",
                "0mm",
                "Radius:=",
                radius,
                "WhichAxis:=",
                "Z",
                "NumSegments:=",
                "0",
            ],
            [
                "NAME:Attributes",
                "Name:=",
                name + "_split",
                "Flags:=",
                "",
                "Color:=",
                "(132 132 193)",
                "Transparency:=",
                0,
                "PartCoordinateSystem:=",
                "Global",
                "UDMId:=",
                "",
                "Materiaobjidue:=",
                '"vacuum"',
                "SolveInside:=",
                True,
            ],
        )

        self.oeditor.Copy(["NAME:Selections", "Selections:=", name])

        self.oeditor.Paste()
        self.oeditor.Intersect(
            ["NAME:Selections", "Selections:=", "{0}1,{0}_split".format(name)],
            ["NAME:IntersectParameters", "KeepOriginals:=", False],
        )

        self.subtract(name, name + "1")
        return True



import os
import string
import random
import time
import datetime
import sys
import traceback
import logging
from functools import wraps
from collections import OrderedDict
import inspect
import itertools

logger = logging.getLogger(__name__)


class MethodNotSupportedError(Exception):
    """ """

    pass


def _write_mes(mes_text, print_on_desktop=False):
    if os.getenv("PYAEDT_SCREEN_LOGS", "True").lower() in ("true", "1", "t"):
        print(mes_text)
    if logger and os.getenv("PYAEDT_FILE_LOGS", "True").lower() in ("true", "1", "t"):
        logger.error(str(mes_text))
    if (
        print_on_desktop
        and os.getenv("PYAEDT_DESKTOP_LOGS", "True").lower() in ("true", "1", "t")
        and "oDesktop" in dir(sys.modules["__main__"])
    ):
        sys.modules["__main__"].oDesktop.AddMessage("", "", 2, (str(mes_text)))


def _exception(ex_info, func, args, kwargs, message="Type Error"):
    """Writes the trace stack to the desktop when a python error occurs

    Parameters
    ----------
    ex_info :

    func :

    args :

    kwargs :

    message :
         (Default value = "Type Error")

    Returns
    -------

    """
    if os.getenv("PYAEDT_SCREEN_LOGS", "True").lower() in ("true", "1", "t"):
        _write_mes("**************************************************************")
        _write_mes("pyaedt Error on Method {}:  {}. Please Check again".format(func.__name__, message), True)
        _write_mes("Arguments Provided: ")

        try:

            if int(sys.version[0]) > 2:
                args_name = list(OrderedDict.fromkeys(inspect.getfullargspec(func)[0] + list(kwargs.keys())))
                args_dict = OrderedDict(list(itertools.zip_longest(args_name, args)) + list(kwargs.items()))
            else:
                args_name = list(OrderedDict.fromkeys(inspect.getargspec(func)[0] + list(kwargs.keys())))
                args_dict = OrderedDict(list(itertools.izip(args_name, args)) + list(kwargs.iteritems()))

            for el in args_dict:
                if el != "self":
                    _write_mes("    {} = {} ".format(el, args_dict[el]))
        except:
            if len(args) > 1:
                _write_mes(args[1:], kwargs)
            else:
                _write_mes(kwargs)
    ex_value = ex_info[1]
    tb_data = ex_info[2]
    tb_trace = traceback.format_tb(tb_data)
    if len(tb_trace) > 1:
        tblist = tb_trace[1].split("\n")
    else:
        tblist = tb_trace[0].split("\n")
    _write_mes(str(ex_value), True)
    for el in tblist:
        # self._main.oDesktop.AddMessage(proj_name, des_name, 2, el)
        if "inner_function" not in el and "**kwargs" not in el:
            _write_mes(el, True)
    _write_mes("")
    _write_mes("")
    _write_mes("Method Docstring: ")
    _write_mes("")
    _write_mes(func.__doc__)
    _write_mes("************************************************************")


def aedt_exception_handler(func):
    """Decorator for pyaedt Exception Management

    Parameters
    ----------
    func :
        method to be decorated

    Returns
    -------
    type
        function return if correctly executed otherwise it will return False and errors will be plotted

    """

    @wraps(func)
    def inner_function(*args, **kwargs):
        if os.getenv("PYAEDT_ERROR_HANDLER", "True").lower() in ("true", "1", "t"):
            try:
                return func(*args, **kwargs)
            except TypeError:
                _exception(sys.exc_info(), func, args, kwargs, "Type Error")
                return False
            except ValueError:
                _exception(sys.exc_info(), func, args, kwargs, "Value Error")
                return False
            except AttributeError:
                _exception(sys.exc_info(), func, args, kwargs, "Attribute Error")
                return False
            except KeyError:
                _exception(sys.exc_info(), func, args, kwargs, "Key Error")
                return False
            except IndexError:
                _exception(sys.exc_info(), func, args, kwargs, "Index Error")
                return False
            except AssertionError:
                _exception(sys.exc_info(), func, args, kwargs, "Assertion Error")
                return False
            except NameError:
                _exception(sys.exc_info(), func, args, kwargs, "Name Error")
                return False
            except IOError:
                _exception(sys.exc_info(), func, args, kwargs, "IO Error")
                return False
            except MethodNotSupportedError:
                message = "This Method is not supported in current AEDT Design Type."
                if os.getenv("PYAEDT_SCREEN_LOGS", "True").lower() in ("true", "1", "t"):
                    print("**************************************************************")
                    print("pyaedt Error on Method {}:  {}. Please Check again".format(func.__name__, message))
                    print("**************************************************************")
                    print("")
                if os.getenv("PYAEDT_FILE_LOGS", "True").lower() in ("true", "1", "t"):
                    logger.error(message)
                return False
            except BaseException:
                _exception(sys.exc_info(), func, args, kwargs, "General or AEDT Error")
                return False
        else:
            return func(*args, **kwargs)

    return inner_function


@aedt_exception_handler
def env_path(input_version):
    """

    Parameters
    ----------
    input_version :


    Returns
    -------

    """
    version = int(input_version[2:4])
    release = int(input_version[5])
    if version < 20:
        if release < 3:
            version -= 1
        else:
            release -= 2
    v_key = "ANSYSEM_ROOT{0}{1}".format(version, release)
    return os.getenv(v_key)


@aedt_exception_handler
def env_value(input_version):
    """

    Parameters
    ----------
    input_version :


    Returns
    -------

    """
    version = int(input_version[2:4])
    release = int(input_version[5])
    if version < 20:
        if release < 3:
            version -= 1
        else:
            release -= 2
    v_key = "ANSYSEM_ROOT{0}{1}".format(version, release)
    return v_key


@aedt_exception_handler
def env_path_student(input_version):
    """Return the Student version Environment Variable value based on an input version string

    Parameters
    ----------
    input_version : str


    Returns
    -------
    str

    Examples
    --------
    >>> env_path_student("2021.1")
    "C:/Program Files/ANSYSEM/ANSYSEM2021.1/Win64"
    """
    version = int(input_version[2:4])
    release = int(input_version[5])
    if version < 20:
        if release < 3:
            version -= 1
        else:
            release -= 2
    v_key = "ANSYSEMSV_ROOT{0}{1}".format(version, release)
    return os.getenv(v_key)


@aedt_exception_handler
def env_value_student(input_version):
    """Return the Student version Environment Variable name based on an input version string

    Parameters
    ----------
    input_version : str


    Returns
    -------
    str

    Examples
    --------
    >>> env_value_student("2021.1")
    "ANSYSEMSV_ROOT211"
    """
    version = int(input_version[2:4])
    release = int(input_version[5])
    if version < 20:
        if release < 3:
            version -= 1
        else:
            release -= 2
    v_key = "ANSYSEMSV_ROOT{0}{1}".format(version, release)
    return v_key


@aedt_exception_handler
def get_filename_without_extension(path):
    """

    Parameters
    ----------
    path :


    Returns
    -------

    """
    return os.path.splitext(os.path.split(path)[1])[0]


@aedt_exception_handler
def generate_unique_name(rootname, suffix="", n=6):
    """Generate a new Random name given a rootname and, optionally a suffix

    Parameters
    ----------
    rootname :
        Root name to which add 6 Random chars
    suffix :
        Suffix to be added (Default value = '')
    n :
        Number of random characters in the name, defaults to 6

    Returns
    -------

    """
    char_set = string.ascii_uppercase + string.digits
    uName = "".join(random.choice(char_set) for _ in range(n))
    unique_name = rootname + "_" + uName
    if suffix:
        unique_name += "_" + suffix
    return unique_name


def retry_ntimes(n, function, *args, **kwargs):
    """

    Parameters
    ----------
    n :

    function :

    *args :

    **kwargs :


    Returns
    -------

    """
    retry = 0
    ret_val = None
    while retry < n:
        try:
            ret_val = function(*args, **kwargs)
            if not ret_val and type(ret_val) is not float and type(ret_val) is not int:
                ret_val = True
        except:
            retry += 1
            time.sleep(0.1)
        else:
            break
    return ret_val


def time_fn(fn, *args, **kwargs):
    start = datetime.datetime.now()
    results = fn(*args, **kwargs)
    end = datetime.datetime.now()
    fn_name = fn.__module__ + "." + fn.__name__
    delta = (end - start).microseconds * 1e-6
    print(fn_name + ": " + str(delta) + "s")
    return results


def isclose(a, b, rel_tol=1e-9, abs_tol=0.0):
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def is_number(a):
    if isinstance(a, float) or isinstance(a, int):
        return True
    elif isinstance(a, str):
        try:
            float(a)
            return True
        except ValueError:
            return False
    else:
        return False
    # return str(a).replace(".", "").replace("+", "").replace("-", "").replace("e","").replace("E","").isnumeric()



import math
import os
import re

from .general_methods import aedt_exception_handler

REAL_IMAG = "RI"
MAG_ANGLE = "MA"
DB_ANGLE = "DB"

keys = {REAL_IMAG: ("real", "imag"), MAG_ANGLE: ("mag", "deg"), DB_ANGLE: ("db20", "deg")}


def parameter(real=None, imag=None, mag=None, db10=None, db20=None, deg=None, rad=None):
    """Initialize a parameter
    Specify:
    * real and (optionally) imag, or
    * mag/db and deg/rad

    Parameters
    ----------
    real :
         (Default value = None)
    imag :
         (Default value = None)
    mag :
         (Default value = None)
    db10 :
         (Default value = None)
    db20 :
         (Default value = None)
    deg :
         (Default value = None)
    rad :
         (Default value = None)

    Returns
    -------

    """
    if real is not None:
        if (mag or db10 or db20 or deg or rad) is not None:
            raise ValueError("Illegal combination of arguments.")
        if imag is None:
            imag = 0
    else:  # real is None
        if (imag is not None) and (mag and db10 and db20 is not None) and (deg and rad is not None):
            raise ValueError("Illegal combination of arguments.")

        if db10 is not None:
            mag = 10 ** (db10 / 10.0)
        elif db20 is not None:
            mag = 10 ** (db20 / 20.0)
        if deg is not None:
            rad = deg * math.pi / 180

        real = mag * math.cos(rad)
        imag = mag * math.sin(rad)

    return complex(real, imag)


def real(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return arg.real


def imag(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return arg.imag


def mag(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return abs(arg)


def db10(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return 10.0 * math.log10(mag(arg))


def db20(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return 20.0 * math.log10(mag(arg))


def rad(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return arg * math.pi / 180


def deg(arg):
    """

    Parameters
    ----------
    arg :


    Returns
    -------
    type


    """
    return arg * 180 / math.pi


formats = {REAL_IMAG: (real, imag), MAG_ANGLE: (mag, deg), DB_ANGLE: (db20, deg)}


class ParseError(Exception):
    """ """

    def __init__(self, message):
        self.message = message

    def __str__(self):
        return "{0}: {1}".format(__class__, self.message)


class TouchstoneData(object):
    """Data Class containing information from Touchstone Read call"""

    def __init__(self, freqs, matrix, portnames):
        self._sweeps_names = ["Freq"]
        self.sweeps = {"Freq": freqs}
        self.ports = portnames
        self.expressions = []
        for el in portnames:
            for el1 in portnames:
                self.expressions.append("S({},{})".format(el, el1))

        self._primary_sweep = "Freq"
        self.solutions_data_real, self.solutions_data_imag = self._solutions_data(matrix)
        self.solutions_data_mag = {}
        self.units_data = {}
        for expr in self.expressions:
            self.solutions_data_mag[expr] = [
                abs(complex(self.solutions_data_real[expr][i], self.solutions_data_imag[expr][i]))
                for i in range(len(self.solutions_data_real[expr]))
            ]

    def _solutions_data(self, matrix):
        """

        Parameters
        ----------
        matrix :


        Returns
        -------

        """
        sols_data_real = {}
        sols_data_imag = {}
        i = 0
        expression = ""
        for element in matrix:
            k = 0
            for expression in self.expressions:
                values = []
                for el in reversed(self._sweeps_names):
                    values.append(self.sweeps[el])
                t = tuple([self.sweeps["Freq"][i]])
                if expression in sols_data_real:
                    sols_data_real[expression].append(element[k].real)
                    sols_data_imag[expression].append(element[k].imag)
                else:
                    sols_data_real[expression] = [element[k].real]
                    sols_data_imag[expression] = [element[k].imag]
                k += 1
            i += 1

        return sols_data_real, sols_data_imag

    def data_magnitude(self, expression=None):
        """Return the data magnitude of the given expression. if no expression is provided, first expression is provided

        Parameters
        ----------
        expression :
            string expression name (Default value = None)

        Returns
        -------
        type
            List of data

        """
        if not expression:
            expression = self.expressions[0]
        return list(self.solutions_data_mag[expression])

    def data_db(self, expression=None):
        """Return the data in db of the given expression. if no expression is provided, first expression is provided

        Parameters
        ----------
        expression :
            string expression name (Default value = None)

        Returns
        -------
        type
            List of data

        """
        if not expression:
            expression = self.expressions[0]
        try:
            return [10 * math.log10(i) for i in self.data_magnitude(expression)]
        except:
            print("Error in DB Computation")
            return None

    def data_real(self, expression=None):
        """Return the real part of data  of the given expression. if no expression is provided, first expression is provided

        Parameters
        ----------
        expression :
            string expression name (Default value = None)

        Returns
        -------
        type
            List of data

        """

        if not expression:
            expression = self.expressions[0]
        return list(self.solutions_data_real[expression])

    def data_imag(self, expression=None):
        """Return the imaginary part of data  of the given expression.

        If no expression is provided, first expression is provided

        Parameters
        ----------
        expression :
            string expression name (Default value = None)

        Returns
        -------
        type
            List of data

        """
        if not expression:
            expression = self.expressions[0]
        return list(self.solutions_data_imag[expression])


@aedt_exception_handler
def get_return_losses(excitation_names, excitation_name_prefix=""):
    """Get the list of all the Returnloss from a list of exctitations.

    If no excitation is provided it will provide a full list of return Losses
    Example: excitation_names ["1","2"] is_touchstone_expression=False output ["S(1,1)",, S(2,2)]
    Example: excitation_names ["S(1,1)","S(1,2)", S(2,2)] is_touchstone_expression=True output ["S(1,1)",, S(2,2)]

    Parameters
    ----------
    excitation_names :
        list of excitation to include
    excitation_name_prefix :
         (Default value = '')

    Returns
    -------
    type
        list of string representing Return Losses of excitations

    """
    spar = []
    if excitation_name_prefix:
        excitation_names = [i for i in excitation_names if excitation_name_prefix.lower() in i.lower()]
    for i in excitation_names:
        spar.append("S({},{})".format(i, i))
    return spar


@aedt_exception_handler
def get_insertion_losses_from_prefix(expressions, tx_prefix, rx_prefix):
    """Get the list of all the Insertion Losses from prefix.

    Parameters
    ----------
    expressions :
        list of Drivers to include or all nets
    reclist :
        list of Receiver to include. Number of Driver = Number of Receiver an
    tx_prefix :
        prefix for TX (eg. "DIE")
    rx_prefix :
        prefix for RX (eg. "BGA")

    Returns
    -------
    type
        list of string representing Insertion Losses of excitations

    """
    spar = []
    left_list = []
    right_list = []
    trlist = [i for i in expressions if tx_prefix in i]
    reclist = [i for i in expressions if rx_prefix in i]
    if len(trlist) != len(reclist):
        print("TX and RX should be same length lists")
        return False
    for i, j in zip(trlist, reclist):
        spar.append("S({},{})".format(i, j))
    return spar


@aedt_exception_handler
def get_insertion_losses_from_lists(txlist, reclist):
    """Get the list of all the Insertion Losses from two list of exctitations (driver and receiver). Optionally prefix can
    be used to retrieve driver and receiver names.
    Example: excitation_names ["1"] ["2"] output ["S(1,2)"]

    Parameters
    ----------
    txlist :
        list of Drivers to include or all nets
    reclist :
        list of Receiver to include. Number of Driver = Number of Receiver an

    Returns
    -------
    type
        list of string representing Insertion Losses of excitations

    """
    spar = []
    if len(txlist) != len(reclist):
        print("TX and RX should be same length lists")
        return False
    for i, j in zip(txlist, reclist):
        spar.append("S({},{})".format(i, j))
    return spar


@aedt_exception_handler
def get_next_xtalk(expressions, tx_prefix=""):
    """Get the list of all the Near End XTalk a list of excitation. Optionally prefix can
    be used to retrieve driver names.
    Example: excitation_names ["1", "2", "3"] output ["S(1,2)", "S(1,3)", "S(2,3)"]

    Parameters
    ----------
    expressions :
        list of Drivers to include
    tx_prefix :
        prefix for TX (eg. "DIE") (Default value = "")

    Returns
    -------
    type
        list of string representing Near End XTalks

    """
    next = []
    if tx_prefix:
        trlist = [i for i in expressions if tx_prefix in i]
    else:
        trlist = expressions
    for i in trlist:
        k = trlist.index(i) + 1
        while k < len(trlist):
            next.append("S({},{})".format(i, trlist[k]))
            k += 1
    return next


@aedt_exception_handler
def get_fext_xtalk_from_list(trlist, reclist, skip_same_index_couples=True):
    """Get the list of all the Far End XTalk from 2 lists of exctitations.

    If skip_same_index_couples is true, the tx and rx with same index
    position will be considered insertion losses and excluded from the list
    Example: excitation_names ["1", "2"] ["3","4"] output ["S(1,4)", "S(2,3)"]

    Parameters
    ----------
    trlist :
        list of Drivers to include
    reclist :
        list of Receiver to include
    skip_same_index_couples :
        Boolean ignore TX and RX couple with same index (Default value = True)

    Returns
    -------
    type
        list of string representing Far End XTalks

    """
    fext = []
    for i in trlist:
        for k in reclist:
            if not skip_same_index_couples or reclist.index(k) != trlist.index(i):
                fext.append("S({},{})".format(i, k))
    return fext


@aedt_exception_handler
def get_fext_xtalk_from_prefix(expressions, tx_prefix, rx_prefix, skip_same_index_couples=True):
    """Get the list of all the Far End XTalk from a list of exctitations and a prefix that will
    be used to retrieve driver and receivers names. If skip_same_index_couples is true, the tx and rx with same index
    position will be considered insertion losses and excluded from the list

    Parameters
    ----------
    expressions :
        list of Receiver to include
    tx_prefix :
        prefix for TX (eg. "DIE")
    reclist :
        list of Receiver to include
    rx_prefix :
        prefix for RX (eg. "BGA")
    skip_same_index_couples :
        Boolean ignore TX and RX couple with same index (Default value = True)

    Returns
    -------
    type
        list of string representing Far End XTalks

    """
    fext = []
    trlist = [i for i in expressions if tx_prefix in i]
    reclist = [i for i in expressions if rx_prefix in i]
    for i in trlist:
        for k in reclist:
            if not skip_same_index_couples or reclist.index(k) != trlist.index(i):
                fext.append("S({},{})".format(i, k))
    return fext


@aedt_exception_handler
def get_worst_curve_from_solution_data(
    solution_data, freq_min=None, freq_max=None, worst_is_higher=True, curve_list=None
):
    """This method analyze a solution data object with multiple curves and find the worst curve returning its name and
     a ordered dictionary with each curve mean. Actual algorithm simply takes the mean of the magnitued over the
     frequency range

    Parameters
    ----------
    solution_data :
        SolutionData or TouchstoneData object
    freq_min :
        minimum frequency to analyze (None to 0) (Default value = None)
    freq_max :
        maximum frequency to analyze (None to max freq) (Default value = None)
    worst_is_higher :
        boolean. if True, the worst curve is the one with higher mean value (Default value = True)
    curve_list :
        list of curves on which to search. None to search on all curves (Default value = None)

    Returns
    -------
    type
        worst element str, dictionary of ordered expression and their mean

    """
    if not curve_list:
        curve_list = solution_data.expressions
    return_loss_freq = solution_data.sweeps["Freq"]
    if not freq_min:
        lower_id = 0
    else:
        lower_id = next(x[0] for x in enumerate(return_loss_freq) if x[1] >= freq_min)
    if not freq_max:
        higher_id = len(return_loss_freq) - 1
    else:
        if freq_max >= return_loss_freq[-1]:
            higher_id = len(return_loss_freq) - 1
        else:
            higher_id = next(x[0] for x in enumerate(return_loss_freq) if x[1] >= freq_max)

    dict_means = {}
    for el in curve_list:
        data1 = solution_data.data_magnitude(el)[lower_id:higher_id]
        mean1 = sum(data1) / len(data1)
        dict_means[el] = mean1
    dict_means = dict(sorted(dict_means.items(), key=lambda item: item[1], reverse=worst_is_higher))
    worst_el = next(iter(dict_means))
    return worst_el, dict_means


def read_touchstone(file_path, verbose=False):
    """Load the contents of a Touchstone file into an NPort

    Parameters
    ----------
    file_path :

    verbose :
         (Default value = False)

    Returns
    -------
    class:`nport.NPort`
        NPort holding data contained in the Touchstone file

    """
    re_filename = re.compile(r"\.s(?P<ports>\d+)+p", re.I)
    m = re_filename.search(file_path)
    ports = int(m.group("ports"))
    if verbose:
        print("File '%s'" % file_path)
        print("  Number of ports (based on file extension) = %d" % ports)
    file_path = os.path.abspath(file_path)
    file = open(file_path, "r")
    (frequnit, type, format, z0) = _parse_option_line(file, verbose)
    freqs = []
    matrices = []
    try:
        while True:
            freq, matrix = _parse_next_sample(file, ports, format)
            freqs.append(freq)
            matrices.append(matrix)
    except EOFError:
        pass
    finally:
        file.close()
    file = open(file_path, "r")
    port_names = _parse_ports_name(file)
    file.close()
    data = TouchstoneData(freqs, matrices, port_names)
    return data


_re_comment = re.compile(r"^\s*!")
_re_options = re.compile(r"^\s*#")
_re_empty = re.compile(r"^\s*$")


def _parse_ports_name(file):
    """Parse and interpret the option line in the touchstone file

    Parameters
    ----------
    file :


    Returns
    -------

    """
    portnames = []
    line = file.readline()
    while not line.startswith("! Port"):
        line = file.readline()
    while line.startswith("! Port"):
        portnames.append(line.split(" = ")[1].strip())
        line = file.readline()
    return portnames


def _parse_option_line(file, verbose=False):
    """Parse and interpret the option line in the touchstone file

    Parameters
    ----------
    file :

    verbose :
         (Default value = False)

    Returns
    -------

    """
    prefix = {"": 1, "K": 1e3, "M": 1e6, "G": 1e9}

    # defaults
    frequnit = 1e9
    format = MAG_ANGLE
    z0 = 50

    # format of the options line (order is unimportant)
    # <frequency unit> <parameter> <format> R <n>
    line = file.readline()
    while not line.startswith("#"):
        line = file.readline()

    options = line[1:].upper().split()

    i = 0
    while i < len(options):
        option = options[i]
        if option in ("GHZ", "MHZ", "KHZ", "HZ"):
            frequnit = prefix[option[:-2]]
        elif option in ("DB", "MA", "RI"):
            format = option
        elif option in ("S", "Y", "Z", "H", "G"):
            parameter = option
        elif option == "R":
            i += 1
            z0 = float(options[i])
        else:
            raise ParseError("unrecognized option: {0}".format(option))
        i += 1

    if verbose:
        print("  Frequency unit: %g Hz" % frequnit)
        print("  Parameter:      %s" % type)
        print("  Format:         %s" % format)
        print("  Reference R:    %g" % z0)

    return (frequnit, type, format, z0)


def _get_next_line_data(file):
    """Returns the data on the next line of the input file as an array
    of floats, skipping comment, option and empty lines.

    Parameters
    ----------
    file :


    Returns
    -------

    """
    line = "!"
    while _re_comment.search(line) or _re_options.search(line) or _re_empty.search(line):
        line = file.readline()
        if not line:  # end of data
            raise EOFError
    data = []
    line = line.split("!")[0]
    for number in line.split():
        data.append(float(number))
    return data


def _parse_next_sample(file, ports, format):
    """Parse the parameters for the next frequency point

    Parameters
    ----------
    file :

    ports :

    format :


    Returns
    -------

    """
    data = _get_next_line_data(file)
    # data lines always contain an even number of values (2 values
    #  per parameter), *unless* a frequency value is included in
    #  front
    assert len(data) % 2 == 1
    freq = data[0]
    data = data[1:]
    count = 0

    port1 = port2 = 1
    list_data = [None for i in range(ports * ports)]

    while True:
        for i in range(int(len(data) / 2)):
            index = 2 * i
            args = {}
            args[keys[format][0]] = data[index]
            args[keys[format][1]] = data[index + 1]
            try:
                if ports == 2:
                    list_data[(port1 - 1) * ports + (port2 - 1)] = parameter(**args)
                else:
                    list_data[(port2 - 1) * ports + (port1 - 1)] = parameter(**args)
            except IndexError:
                raise Exception("more ports than reported in the file " "extension")

            port1 += 1
            if port1 > ports:
                port2 += 1
                port1 = 1

        count += len(data) / 2
        if port2 > ports:
            break
        data = _get_next_line_data(file)
        if len(data) % 2 != 0:
            raise Exception("less ports than reported in the file extension")

    assert count == ports ** 2
    return (freq, list_data)



"""
This module contains these classes: `CSVDataset`, `DataSet`, `Expression`, `Variable`, and `VariableManager`.

This module is used to create and edit design and project variables in the 3D tools.

Examples
--------
>>> from pyaedt import Hfss
>>> hfss = Hfss()
>>> hfss["$d"] = "5mm"
>>> hfss["d"] = "5mm"
>>> hfss["postd"] = "1W"

"""
from __future__ import absolute_import

import math
import os
import re

from .. import aedt_exception_handler
from ..generic.general_methods import is_number


@aedt_exception_handler
def dB(x, inverse=True):
    """Convert db to decimal"""
    if inverse:
        return 20 * math.log10(x)
    else:
        return math.pow(10, x / 20.0)


@aedt_exception_handler
def fah2kel(val, inverse=True):
    """Convert a temperature from Fahrenheit to Kelvin.

    Parameters
    ----------
    val : float
        Temperature value in Fahrenheit.
    inverse : bool, optional
        The default is ``True``.

    Returns
    -------
    float
        Temperature value converted to Kelvin.

    """
    if inverse:
        return (val - 273.15) * 9 / 5 + 32
    else:
        return (val - 32) * 5 / 9 + 273.15


@aedt_exception_handler
def cel2kel(val, inverse=True):
    """Convert a temperature from Celsius to Kelvin.

    Parameters
    ----------
    val : float
        Temperature value in Celsius.
    inverse : bool, optional
        The default is ``True``.

    Returns
    -------
    float
        Temperature value converted to Kelvin.

    """

    if inverse:
        return val - 273.15
    else:
        return val + 273.15


@aedt_exception_handler
def unit_system(units):
    """Retrieve the name of the unit system associated with a unit string.

    Parameters
    ----------
    units : str
        Units for retrieving the associated unit system name.

    Returns
    -------
    str
        Key from the ``AEDT_units`` when successful. For example, ``"AngularSpeed"``.
    ``False`` when the units specified are not defined in AEDT units.

    """
    found = False
    for unit_type, unit_list in AEDT_units.items():
        for test_unit in unit_list:
            if test_unit == units:
                found = True
                break
        if found:
            return unit_type
    return False


# TODO Add additional units
rad2deg = 180.0 / math.pi
deg2rad = math.pi / 180
hour2sec = 3600.0
min2sec = 60.0
sec2min = 1 / 60.0
sec2hour = 1 / 3600.0
inv2pi = 0.5 / math.pi
v2pi = 2.0 * math.pi
m2in = 0.0254
m2miles = 1609.344051499

# Value in new units equals value in old units * scaling vector of new units
# e.g. value_in_deg Variable("1rad") = 1 * rad2deg
AEDT_units = {
    "AngularSpeed": {
        "deg_per_hr": hour2sec * deg2rad,
        "rad_per_hr": hour2sec,
        "deg_per_min": min2sec * deg2rad,
        "rad_per_min": min2sec,
        "deg_per_sec": deg2rad,
        "rad_per_sec": 1.0,
        "rev_per_sec": v2pi,
        "per_sec": v2pi,
        "rpm": sec2min * v2pi,
    },
    "Angle": {"deg": deg2rad, "rad": 1.0, "degmin": deg2rad * sec2min, "degsec": deg2rad * sec2hour},
    "Current": {
        "fA": 1e-15,
        "pA": 1e-12,
        "nA": 1e-9,
        "uA": 1e-6,
        "mA": 1e-3,
        "A": 1.0,
        "kA": 1e3,
        "MegA": 1e6,
        "gA": 1e9,
        "dBA": (dB,),
    },
    "Flux": {"Wb": 1.0, "mx": 1e-8, "vh": 3600, "vs": 1.0},
    "Freq": {"Hz": 1.0, "kHz": 1e3, "MHz": 1e6, "GHz": 1e9, "THz": 1e12, "rps": 1.0, "per_sec": 1.0},
    "Inductance": {"fH": 1e-15, "pH": 1e-12, "nH": 1e-9, "uH": 1e-6, "mH": 1e-3, "H": 1.0},
    "Length": {
        "fm": 1e-15,
        "pm": 1e-12,
        "nm": 1e-9,
        "um": 1e-6,
        "mm": 1e-3,
        "cm": 1e-2,
        "dm": 1e-1,
        "meter": 1.0,
        "km": 1e3,
        "uin": m2in * 1e-6,
        "mil": m2in * 1e-3,
        "in": m2in,
        "ft": m2in * 12,
        "yd": m2in * 144,
    },
    "Mass": {"ug": 1e-9, "mg": 1e-6, "g": 1e-3, "kg": 1.0, "ton": 1000, "oz": 0.0283495, "lb": 0.453592},
    "None": {
        "f": 1e-15,
        "p": 1e-12,
        "n": 1e-9,
        "u": 1e-6,
        "m": 1e-3,
        "": 1.0,
        "k": 1e3,
        "meg": 1e6,
        "g": 1e9,
        "t": 1e12,
    },
    "Resistance": {"uOhm": 1e-6, "mOhm": 1e-3, "ohm": 1.0, "kOhm": 1e3, "megohm": 1e6, "GOhm": 1e9},
    "Speed": {
        "mm_per_sec": 1e-3,
        "cm_per_sec": 1e-2,
        "m_per_sec": 1.0,
        "km_per_sec": 1e3,
        "inches_per_sec": m2in,
        "feet_per_sec": m2in * 12,
        "feet_per_min": m2in * 12 * sec2min,
        "km_per_min": 60e3,
        "m_per_h": 3600,
        "miles_per_hour": m2miles * sec2hour,
        "miles_per_minute": m2miles * sec2min,
        "miles_per_sec": m2miles,
    },
    "Time": {
        "fs": 1e-15,
        "ps": 1e-12,
        "ns": 1e-9,
        "us": 1e-6,
        "ms": 1e-3,
        "s": 1,
        "min": 60,
        "hour": 3600,
        "day": 3600 * 12,
    },
    "Torque": {
        "fNewtonMeter": 1e-15,
        "pNewtonMeter": 1e-12,
        "nNewtonMeter": 1e-9,
        "uNewtonMeter": 1e-6,
        "mNewtonMeter": 1e-3,
        "cNewtonMeter": 1e-2,
        "NewtonMeter": 1,
        "kNewtonMeter": 1e3,
        "megNewtonMeter": 1e6,
        "gNewtonMeter": 1e9,
    },
    "Voltage": {
        "fV": 1e-15,
        "pV": 1e-12,
        "nV": 1e-9,
        "uV": 1e-6,
        "mV": 1e-3,
        "V": 1.0,
        "kV": 1e3,
        "MegV": 1e6,
        "gV": 1e9,
        "dBV": (dB,),
    },
    "Temperature": {"kel": 1.0, "cel": (cel2kel,), "fah": (fah2kel,)},
    "Power": {
        "fW": 1e-15,
        "pW": 1e-12,
        "nW": 1e-9,
        "uW": 1e-6,
        "mW": 1e-3,
        "W": 1.0,
        "kW": 1e3,
        "megW": 1e6,
        "gW": 1e9,
    },
}
SI_units = {
    "AngularSpeed": "rad_per_sec",
    "Angle": "rad",
    "Current": "A",
    "Flux": "vs",
    "Freq": "Hz",
    "Inductance": "H",
    "Length": "meter",
    "Mass": "kg",
    "None": "",
    "Resistance": "ohm",
    "Time": "s",
    "Torque": "NewtonMeter",
    "Voltage": "V",
    "Temperature": "kel",
    "Power": "W",
}

unit_system_operations = {
    # Multiplication of physical domains
    "Voltage_multiply_Current": "Power",
    "Torque_multiply_AngularSpeed": "Power",
    "AngularSpeed_multiply_Time": "Angle",
    "Current_multiply_Resistance": "Voltage",
    "AngularSpeed_multiply_Inductance": "Resistance",
    "Speed_multiply_Time": "Length",
    # Division of Physical Domains
    "Power_divide_Voltage": "Current",
    "Power_divide_Current": "Voltage",
    "Power_divide_AngularSpeed": "Torque",
    "Power_divide_Torque": "AngularSpeed",
    "Angle_divide_AngularSpeed": "Time",
    "Angle_divide_Time": "AngularSpeed",
    "Voltage_divide_Current": "Resistance",
    "Voltage_divide_Resistance": "Current",
    "Resistance_divide_AngularSpeed": "Inductance",
    "Resistance_divide_Inductance": "AngularSpeed",
    "None_divide_Freq": "Time",
    "None_divide_Time": "Freq",
    "Length_divide_Time": "Speed",
    "Length_divide_Speed": "Time",
}


def _resolve_unit_system(unit_system_1, unit_system_2, operation):
    """Retrieve the unit string of an arithmetic operation on ``Variable`` objects. If no resulting unit system
    is defined for a specific operation (in unit_system_operations), an empty string is returned

    Parameters
    ----------
    unit_system_1 : str
        Name of a unit system, which is a key of ``AEDT_units``.
    unit_system_2 : str
        Name of another unit system, which is a key of ``AEDT_units``.
    operation : str
        Name of an operator within the data set of ``["multiply" ,"divide"]``.

    Returns
    -------
    str
        Unit system when successful, ``""`` when failed.

    """
    try:
        key = "{}_{}_{}".format(unit_system_1, operation, unit_system_2)
        result_unit_system = unit_system_operations[key]
        return SI_units[result_unit_system]
    except KeyError:
        return ""


class CSVDataset:
    """Reads in a CSV file and extracts data, which can be augmented with constant values.

    Parameters
    ----------
    csv_file: str, optional
        Input file consisting of delimited data with the first line as the header.
        The CSV value includes the header and data, which supports AEDT units information
        such as ``"1.23Wb"``. You can also augment the data with constant values.
    separator: str, optional
        Value to use for the delimiter. The default is``None`` in which case a comma is
        assumed.
    units_dict: dict, optional
        Dictionary consisting of ``{Variable Name: unit}`` to rescale the data
        if it is not in the desired unit system.
    append_dict: dict, optional
        Dictionary consisting of ``{New Variable Name: value}`` to add variables
        with constant values to all data points. This dictionary is used to add
        multiple sweeps to one result file.
    valid_solutions : bool, optional
        The default is ``True``.
    invalid_solutions : bool, optional
        The default is ``False``.

    """

    @property
    def number_of_rows(self):
        """Number of rows."""
        if self._data:
            for variable, data_list in self._data.items():
                return len(data_list)
        else:
            return 0

    @property
    def number_of_columns(self):
        """Number of columns."""
        return len(self._header)

    @property
    def header(self):
        """Header."""
        return self._header

    @property
    def data(self):
        """Data."""
        return self._data

    @property
    def path(self):
        """Path."""
        return os.path.dirname(os.path.realpath(self._csv_file))

    def __init__(
        self,
        csv_file=None,
        separator=None,
        units_dict=None,
        append_dict=None,
        valid_solutions=True,
        invalid_solutions=False,
    ):

        self._header = []
        self._data = {}
        self._unit_dict = {}
        self._append_dict = {}

        # Set the index counter explicitly to zero
        self._index = 0

        if separator:
            self._separator = separator
        else:
            self._separator = ","

        if units_dict:
            self._unit_dict = units_dict

        if append_dict:
            self._append_dict = append_dict

        self._csv_file = csv_file
        if csv_file:
            with open(csv_file, "r") as fi:
                file_data = fi.readlines()
                for line in file_data:
                    if self._header:
                        line_data = line.strip().split(self._separator)
                        # Check for invalid data in the line (fields with 'nan')
                        if not "nan" in line_data:
                            for j, value in enumerate(line_data):
                                var_name = self._header[j]
                                if var_name in self._unit_dict:
                                    var_value = Variable(value).rescale_to(self._unit_dict[var_name]).numeric_value
                                else:
                                    var_value = Variable(value).value
                                self._data[var_name].append(var_value)

                            # Add augmented quantities
                            for entry in self._append_dict:
                                var_value_str = self._append_dict[entry]
                                numeric_value = Variable(var_value_str).numeric_value
                                self._data[entry].append(numeric_value)

                    else:
                        self._header = line.strip().split(",")
                        for additional_quantity_name in self._append_dict:
                            self._header.append(additional_quantity_name)
                        for quantity_name in self._header:
                            self._data[quantity_name] = []

        pass

    @aedt_exception_handler
    def __getitem__(self, item):
        variable_list = item.split(",")
        data_out = CSVDataset()
        for variable in variable_list:
            found_variable = False
            for key_string in self._data:
                if variable in key_string:
                    found_variable = True
                    break
            assert found_variable, "Input string {} is not a key of the data dictionary.".format(variable)
            data_out._data[variable] = self._data[key_string]
            data_out._header.append(variable)
        return data_out

    @aedt_exception_handler
    def __add__(self, other):
        assert self.number_of_columns == other.number_of_columns, "Inconsistent number of columns"
        # Create a new object to return, avoiding changing the original inputs
        new_dataset = CSVDataset()
        # Add empty columns to new_dataset
        for column in self._data:
            new_dataset._data[column] = []

        # Add the data from 'self' to a the new dataset
        for column, row_data in self.data.items():
            for value in row_data:
                new_dataset._data[column].append(value)

        # Add the data from 'other' to a the new dataset
        for column, row_data in other.data.items():
            for value in row_data:
                new_dataset._data[column].append(value)

        return new_dataset

    def __iadd__(self, other):
        """Incrementally add the dataset in one CSV file to a dataset in another CSV file.

        .. note:
           This assumes that the number of columns in both datasets are the same,
           or that one of the datasets is empty. No checking is done for
           equivalency of units or variable names.

        """

        # Handle the case of an empty data set and create empty lists for the column data
        if self.number_of_columns == 0:
            self._header = other.header
            for column in other.data:
                self._data[column] = []

        assert self.number_of_columns == other.number_of_columns, "Inconsistent number of columns"

        # Append the data from 'other'
        for column, row_data in other.data.items():
            for value in row_data:
                self._data[column].append(value)

        return self

    # Called when iteration is initialized
    def __iter__(self):

        self._index = 0
        return self

    # Create an iterator to yield the row data as a string as we loop through the object
    def __next__(self):
        if self._index < (self.number_of_rows - 1):
            output = []
            for column in self._header:
                string_value = str(self._data[column][self._index])
                output.append(string_value)
            output_string = " ".join(output)
            self._index += 1
        else:
            raise StopIteration

        return output_string

    def next(self):
        """Yield the next row."""
        return self.__next__()


@aedt_exception_handler
def _find_units_in_dependent_variables(variable_value, full_variables={}):
    m2 = re.findall(r"[0-9.]+ *([a-z_A-Z]+)", variable_value)
    if len(m2) > 0:
        if len(set(m2)) <= 1:
            return m2[0]
        else:
            if unit_system(m2[0]):
                return SI_units[unit_system(m2[0])]
    else:
        m1 = re.findall(r"(?<=[/+-/*//^/(/[])([a-z_A-Z]\w*)", variable_value.replace(" ", ""))
        m2 = re.findall(r"^([a-z_A-Z]\w*)", variable_value.replace(" ", ""))
        m = list(set(m1).union(m2))
        for i, v in full_variables.items():
            if i in m and _find_units_in_dependent_variables(v):
                return _find_units_in_dependent_variables(v)
    return ""


@aedt_exception_handler
def decompose_variable_value(variable_value, full_variables={}):
    """Decompose a variable value.

    Parameters
    ----------
    variable_value : float
    full_variables : dict
    Returns
    -------

    """
    # set default return values - then check for valid units
    float_value = variable_value
    units = ""

    if is_number(variable_value):
        float_value = float(variable_value)
    elif isinstance(variable_value, str) and variable_value != "nan":
        try:
            # Handle a numerical value in string form
            float_value = float(variable_value)
        except ValueError:
            # search for a valid units string at the end of the variable_value
            loc = re.search("[a-z_A-Z]+$", variable_value)
            units = _find_units_in_dependent_variables(variable_value, full_variables)

            if loc:
                loc_units = loc.span()[0]
                extract_units = variable_value[loc_units:]
                if unit_system(extract_units):
                    try:
                        float_value = float(variable_value[0:loc_units])
                        units = extract_units
                    except ValueError:
                        float_value = variable_value

    return float_value, units


class VariableManager(object):
    """Manages design properties and project variables.

    Design properties are the local variables in a design. Project
    variables are defined at the project level and start with ``$``.

    This class provides access to all variables or a subset of the
    variables. Manipulation of the numerical or string definitions of
    variable values is provided in the
    :class:`pyaedt.application.Variables.Variable` class.

    Parameters
    ----------
    variables : dict
        Dictionary of all design properties and project variables in
        the active design.
    design_variables : dict
        Dictionary of all design properties in the active design.
    project_variables : dict
        Dictionary of all project variables available to the active
        design (key by variable name).
    dependent_variables : dict
        Dictionary of all dependent variables available to the active
        design (key by variable name).
    independent_variables : dict
       Dictionary of all independent variables (constant numeric
       values) available to the active design (key by variable name).
    independent_design_variables : dict

    independent_project_variables : dict

    variable_names : str or list
        One or more variable names.
    project_variable_names : str or list
        One or more project variable names.
    design_variable_names : str or list
        One or more design variable names.
    dependent_variable_names : str or list
        All dependent variable names within the project.
    independent_variable_names : list of str
        All independent variable names within the project. These can
        be sweep variables for optimetrics.
    independent_project_variable_names : str or list
        All independent project variable names within the
        project. These can be sweep variables for optimetrics.
    independent_design_variable_names : str or list
        All independent design properties (local variables) within the
        project. These can be sweep variables for optimetrics.

    See Also
    --------
    pyaedt.application.Variables.Variable

    Examples
    --------

    >>> from pyaedt.maxwell import Maxwell3d
    >>> from pyaedt.desktop import Desktop
    >>> d = Desktop()
    >>> aedtapp = Maxwell3d()

    Define some test variables.

    >>> aedtapp["Var1"] = 3
    >>> aedtapp["Var2"] = "12deg"
    >>> aedtapp["Var3"] = "Var1 * Var2"
    >>> aedtapp["$PrjVar1"] = "pi"

    Get the variable manager for the active design.

    >>> v = aedtapp.variable_manager

    Get a dictionary of all project and design variables.

    >>> v.variables
    {'Var1': <pyaedt.application.Variables.Variable at 0x2661f34c448>,
     'Var2': <pyaedt.application.Variables.Variable at 0x2661f34c308>,
     'Var3': <pyaedt.application.Variables.Expression at 0x2661f34cb48>,
     '$PrjVar1': <pyaedt.application.Variables.Expression at 0x2661f34cc48>}

    Get a dictionary of only the design variables.

    >>> v.design_variables
    {'Var1': <pyaedt.application.Variables.Variable at 0x2661f339508>,
     'Var2': <pyaedt.application.Variables.Variable at 0x2661f3415c8>,
     'Var3': <pyaedt.application.Variables.Expression at 0x2661f341808>}

    Get a dictionary of only the independent design variables.

    >>> v.independent_design_variables
    {'Var1': <pyaedt.application.Variables.Variable at 0x2661f335d08>,
     'Var2': <pyaedt.application.Variables.Variable at 0x2661f3557c8>}

    """

    @property
    def variables(self):
        """Variables.

        Returns
        -------
        dict
            Dictionary of the `Variable` objects for each project variable and each
            design property in the active design.

        """
        return self._variable_dict([self.odesign, self.oproject])

    @property
    def design_variables(self):
        """Design variables.

        Returns
        -------
        dict
            Dictionary of the design properties (local properties) in the design.
        """
        return self._variable_dict([self.odesign])

    @property
    def project_variables(self):
        """Project variables.

        Returns
        -------
        dict
            Dictionary of the project properties.
        """
        return self._variable_dict([self.oproject])

    @property
    def independent_variables(self):
        """Independent variables.

        Returns
        -------
        dict
            Dictionary of the independent variables (constant numeric
            values) available to the design.
        """
        return self._variable_dict([self.odesign, self.oproject], dependent=False)

    @property
    def independent_project_variables(self):
        """Independent project variables.

        Returns
        -------
        dict
            Dictionary of the independent project variables available to the design.
        """
        return self._variable_dict([self.oproject], dependent=False)

    @property
    def independent_design_variables(self):
        """Independent design variables.

        Returns
        -------
        dict
            Dictionary of the independent design properties (local
            variables) available to the design.
        """
        return self._variable_dict([self.odesign], dependent=False)

    @property
    def dependent_variables(self):
        """Dependent variables.

        Returns
        -------
        dict
            Dictionary of the dependent design properties (local
            variables) and project variables available to the design.

        """
        return self._variable_dict([self.odesign, self.oproject], independent=False)

    @property
    def variable_names(self):
        """List of variables."""
        return [var_name for var_name in self.variables]

    @property
    def project_variable_names(self):
        """List of project variables."""
        return [var_name for var_name in self.project_variables]

    @property
    def independent_project_variable_names(self):
        """List of independent project variables."""
        return [var_name for var_name in self.independent_project_variables]

    @property
    def design_variable_names(self):
        """List of design variables."""
        return [var_name for var_name in self.design_variables]

    @property
    def independent_design_variable_names(self):
        """List of independent design variables."""
        return [var_name for var_name in self.independent_design_variables]

    @property
    def independent_variable_names(self):
        """List of independent variables."""
        return [var_name for var_name in self.independent_variables]

    @property
    def dependent_variable_names(self):
        """List of dependent variables."""
        return [var_name for var_name in self.dependent_variables]

    @property
    def oproject(self):
        """Project."""
        return self._parent._oproject

    @property
    def odesign(self):
        """Design."""
        return self._parent._odesign

    @property
    def _messenger(self):
        """Messenger."""
        return self._parent._messenger

    def __init__(self, parent):
        # Global Desktop Environment
        self._parent = parent

    @aedt_exception_handler
    def __delitem__(self, key):
        """Implement del with array name or index."""
        self.delete_variable(key)

    @aedt_exception_handler
    def __getitem__(self, variable_name):
        return self.variables[variable_name]

    @aedt_exception_handler
    def __setitem__(self, variable, value):
        self.set_variable(variable, value)
        return True

    @aedt_exception_handler
    def _variable_dict(self, object_list, dependent=True, independent=True):
        """Retrieve the variable dictionary.

        Parameters
        ----------
        object_list : list
            List of objects.
        dependent : bool, optional
             Whether to include dependent variables. The default is ``True``.
        independent : bool, optional
             Whether to include independent variables. The default is ``True``.

        Returns
        -------
        dict
            Dictionary of the specified variables.

        """
        var_dict = {}
        all_names = {}
        for obj in object_list:
            for variable_name in obj.GetVariables():
                variable_expression = self.get_expression(variable_name)
                all_names[variable_name] = variable_expression
                try:
                    value = Variable(variable_expression)
                    if independent and is_number(value.value):
                        var_dict[variable_name] = value
                    elif dependent and type(value.value) is str:
                        float_value = self._parent.get_evaluated_value(variable_name)
                        var_dict[variable_name] = Expression(variable_expression, float_value, all_names)
                except:
                    if dependent:
                        float_value = self._parent.get_evaluated_value(variable_name)
                        var_dict[variable_name] = Expression(variable_expression, float_value, all_names)
        return var_dict

    @aedt_exception_handler
    def get_expression(self, variable_name):
        """Retrieve the variable value of a project or design variable as a string."""
        return self.aedt_object(variable_name).GetVariableValue(variable_name)

    @aedt_exception_handler
    def aedt_object(self, variable):
        """Retrieve an AEDT object.

        Parameters
        ----------
        variable : str
        Name of the variable.

        """
        if variable[0] == "$":
            return self.oproject
        else:
            return self.odesign

    @aedt_exception_handler
    def set_variable(
        self, variable_name, expression=None, readonly=False, hidden=False, description=None, overwrite=True
    ):
        """Set the value of a design property or project variable.

        Parameters
        ----------
        variable_name : str
            Name of the design property or project variable
            (``$var``). If this variable does not exist, a new one is
            created and a value is set.
        expression : str
            Valid string expression within the AEDT design and project
            structure.  For example, ``"3*cos(34deg)"``.
        readonly : bool, optional
            Whether to set the design property or project variable to
            read-only. The default is ``False``.
        hidden :  bool, optional
            Whether to hide the design property or project variable. The
            default is ``False``.
        description : str, optional
            Text to display for the design property or project variable in the
            ``Properties`` window. The default is ``None``.
        overwrite : bool, optional
            Whether to overwrite an existing value for the design
            property or project variable. The default is ``False``, in
            which case this method is ignored.

        Returns
        -------
        bool
             ``True`` when successful, ``False`` when failed.

        Examples
        --------
        Set the value of design property ``p1`` to ``"10mm"``,
        creating the property if it does not already eixst.

        >>> aedtapp.variable_manager.set_variable("p1", expression="10mm")

        Set the value of design property ``p1`` to ``"20mm"`` only if
        the property does not already exist.

        >>> aedtapp.variable_manager.set_variable("p1", expression="20mm", overwrite=False)

        Set the value of design property ``p2`` to ``"10mm"``,
        creating the property if it does not already exist. Also make
        it read-only and hidden and add a description.

        >>> aedtapp.variable_manager.set_variable(variable_name="p2", expression="10mm", readonly=True, hidden=True,
        ...                                       description="This is the description of this variable.")

        Set the value of the project variable ``$p1`` to ``"30mm"``,
        creating the variable if it does not exist.

        >>> aedtapp.variable_manager.set_variable["$p1"] == "30mm"

        """
        if not description:
            description = ""

        desktop_object = self.aedt_object(variable_name)
        test = desktop_object.GetName()
        proj_name = self.oproject.GetName()
        var_type = "Project" if test == proj_name else "Local"

        prop_type = "VariableProp"

        if isinstance(expression, str):
            # Handle string type variable (including arbitrary expression)# Handle input type variable
            variable = expression
        elif isinstance(expression, Variable):
            # Handle input type variable
            variable = expression.string_value
        elif is_number(expression):
            # Handle input type int/float, etc (including numeric 0)
            variable = str(expression)
        # Handle None, "" as Separator
        elif isinstance(expression, list):
            variable = str(expression)
        elif not expression:
            prop_type = "SeparatorProp"
            variable = ""
            try:
                if self.delete_separator(variable_name):
                    desktop_object.Undo()
                    self._messenger.clear_messages()
                    return
            except:
                pass
        else:
            raise Exception("Unhandled input type to the design property or project variable.")

        if "post" in variable_name.lower()[0:5]:
            prop_type = "PostProcessingVariableProp"

        # Get all design and project variables in lower case for a case-sensitive comparison
        var_list = desktop_object.GetVariables()
        lower_case_vars = [var_name.lower() for var_name in var_list]

        if variable_name.lower() not in lower_case_vars:

            desktop_object.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:{0}VariableTab".format(var_type),
                        ["NAME:PropServers", "{0}Variables".format(var_type)],
                        [
                            "NAME:NewProps",
                            [
                                "NAME:" + variable_name,
                                "PropType:=",
                                prop_type,
                                "UserDef:=",
                                True,
                                "Value:=",
                                variable,
                                "Description:=",
                                description,
                                "ReadOnly:=",
                                readonly,
                                "Hidden:=",
                                hidden,
                            ],
                        ],
                    ],
                ]
            )
        elif overwrite:
            desktop_object.ChangeProperty(
                [
                    "NAME:AllTabs",
                    [
                        "NAME:{}VariableTab".format(var_type),
                        ["NAME:PropServers", "{}Variables".format(var_type)],
                        [
                            "NAME:ChangedProps",
                            [
                                "NAME:" + variable_name,
                                "Value:=",
                                variable,
                                "Description:=",
                                description,
                                "ReadOnly:=",
                                readonly,
                                "Hidden:=",
                                hidden,
                            ],
                        ],
                    ],
                ]
            )

        return True

    @aedt_exception_handler
    def delete_separator(self, separator_name):
        """Delete a separator from either the active project or design.

        Parameters
        ----------
        separator_name : str
            Value to use for the delimiter.

        Returns
        -------
        bool
            ``True`` when the separator exists and can be deleted, ``False`` otherwise.

        """
        object_list = [(self.odesign, "Local"), (self.oproject, "Project")]

        for object_tuple in object_list:
            desktop_object = object_tuple[0]
            var_type = object_tuple[1]
            try:
                desktop_object.ChangeProperty(
                    [
                        "NAME:AllTabs",
                        [
                            "NAME:{0}VariableTab".format(var_type),
                            ["NAME:PropServers", "{0}Variables".format(var_type)],
                            ["NAME:DeletedProps", separator_name],
                        ],
                    ]
                )
                return True
            except:
                pass
        return False

    @aedt_exception_handler
    def delete_variable(self, sVarName):
        """Delete a variable.

        Parameters
        ----------
        sVarName : str
            Name of the variable.


        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        desktop_object = self.aedt_object(sVarName)
        var_type = "Project" if desktop_object == self.oproject else "Local"

        var_list = desktop_object.GetVariables()
        lower_case_vars = [var_name.lower() for var_name in var_list]

        if sVarName.lower() in lower_case_vars:
            try:
                desktop_object.ChangeProperty(
                    [
                        "NAME:AllTabs",
                        [
                            "NAME:{0}VariableTab".format(var_type),
                            ["NAME:PropServers", "{0}Variables".format(var_type)],
                            ["NAME:DeletedProps", sVarName],
                        ],
                    ]
                )
                return True
            except:
                pass
        return False


class Variable(object):
    """Stores design properties and project variables and provides operations to perform on them.

    Parameters
    ----------
    value : float
        Numerical value of the variable in SI units.
    units : str
        Units for the value.

    Examples
    --------

    >>> from pyaedt.application.Variables import Variable

    Define a variable using a string value consistent with the AEDT properties.

    >>> v = Variable("45mm")

    Define an unitless variable with a value of 3.0.

    >>> v = Variable(3.0)

    Define a variable defined by a numeric result and a unit string.

    >>> v = Variable(3.0 * 4.5, units="mm")
    >>> assert v.numeric_value = 13.5
    >>> assert v.units = "mm"

    """

    def __init__(self, value, units=None):

        if units:
            if unit_system(units):
                specified_units = units

        self._units = None
        self._expression = value
        self._value, self._units = decompose_variable_value(value)

        # If units have been specified, check for a conflict and otherwise use the specified unit system
        if units:
            assert not self._units, "The unit specification {} is inconsistent with the identified units {}.".format(
                specified_units, self._units
            )
            self._units = specified_units

        if is_number(self._value):
            scale = AEDT_units[self.unit_system][self._units]
            if isinstance(scale, tuple):
                self._value = scale[0](self._value, inverse=False)
            else:
                self._value = self._value * scale

    @property
    def unit_system(self):
        """Unit system of the expression as a string."""
        return unit_system(self._units)

    @property
    def units(self):
        """Units."""
        return self._units

    @property
    def value(self):
        """Value."""
        return self._value

    @property
    def numeric_value(self):
        """Numeric part of the expression as a float value."""
        if is_number(self._value):
            scale = AEDT_units[self.unit_system][self._units]
        if isinstance(scale, tuple):
            return scale[0](self._value, True)
        else:
            return self._value / scale

    @property
    def string_value(self):
        """String value.

        The numeric value with the unit is concatenated and returned as a string. The numeric display
        in the modeler and the string value can differ. For example, you might see ``10mm`` in the
        modeler and see ``10.0mm`` returned as the string value.

        """
        return ("{}{}").format(self.numeric_value, self._units)

    @aedt_exception_handler
    def rescale_to(self, units):
        """Rescale the expression to a new unit within the current unit system.

        Parameters
        ----------
        units : str
            Units to rescale to.

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        >>> v = Variable("10W")
        >>> assert v.numeric_value == 10
        >>> assert v.units == "W"
        >>> v.rescale_to("kW")
        >>> assert v.numeric_value == 0.01
        >>> assert v.units == "kW"

        """
        new_unit_system = unit_system(units)
        assert (
            new_unit_system == self.unit_system
        ), "New unit system {0} is inconsistent with the current unit system {1}."
        self._units = units
        return self

    @aedt_exception_handler
    def format(self, format):
        """Retrieve the string value with the specified numerical formatting.

        Parameters
        ----------
        format : str
            Format for the numeric value of the string. For example, ``'06.2f'``. For
            more information, see the `PyFormat documentation <https://pyformat.info/>`_.

        Returns
        -------
        str
            String value with the specified numerical formatting.

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        >>> v = Variable("10W")
        >>> assert v.format("f") == '10.000000W'
        >>> assert v.format("06.2f") == '010.00W'
        >>> assert v.format("6.2f") == ' 10.00W'

        """
        return ("{0:" + format + "}{1}").format(self.numeric_value, self._units)

    @aedt_exception_handler
    def __mul__(self, other):
        """Multiply the variable with a number or another variable and return a new object.

        Parameters
        ---------
        other : numbers.Number or variable
            Object to be multiplied.

        Returns
        -------
        type
            Variable.

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        Multiply ``'Length1'`` by unitless ``'None'``` to obtain ``'Length'``.
        A numerical value is also considered to be unitless.

        >>> v1 = Variable("10mm")
        >>> v2 = Variable(3)
        >>> result_1 = v1 * v2
        >>> result_2 = v1 * 3
        >>> assert result_1.numeric_value == 30.0
        >>> assert result_1.unit_system == "Length"
        >>> assert result_2.numeric_value == result_1.numeric_value
        >>> assert result_2.unit_system == result_1.unit_system

        Multiply voltage times current to obtain power.

        >>> v3 = Variable("3mA")
        >>> v4 = Variable("40V")
        >>> result_3 = v3 * v4
        >>> assert result_3.numeric_value == 0.12
        >>> assert result_3.units == "W"
        >>> assert result_3.unit_system == "Power"

        """
        assert is_number(other) or isinstance(other, Variable), "Multiplier must be a scalar quantity or a variable."
        if is_number(other):
            result_value = self.numeric_value * other
            result_units = self.units
        else:
            if self.unit_system == "None":
                return self.numeric_value * other
            elif other.unit_system == "None":
                return other.numeric_value * self
            else:
                result_value = self.value * other.value
                result_units = _resolve_unit_system(self.unit_system, other.unit_system, "multiply")
                if not result_units:
                    result_units = _resolve_unit_system(other.unit_system, self.unit_system, "multiply")

        return Variable("{}{}".format(result_value, result_units))

    __rmul__ = __mul__

    @aedt_exception_handler
    def __add__(self, other):
        """Add the variable to another variable to return a new object.

        Parameters
        ---------
        other : Variable
            Object to be multiplied.

        Returns
        -------
        type
            Variable.

        Examples
        --------
        >>> from pyaedt.application.Variables import Variable

        >>> v1 = Variable("3mA")
        >>> v2 = Variable("10A")
        >>> result = v1 + v2
        >>> assert result.numeric_value == 10.003
        >>> assert result.units == "A"
        >>> assert result.unit_system == "Current"

        """
        assert isinstance(other, Variable), "You can only add a variable with another variable."
        assert (
            self.unit_system == other.unit_system
        ), "Only ``Variable`` objects with the same unit system can be added."
        result_value = self.value + other.value
        result_units = SI_units[self.unit_system]
        # If the units of the two operands are different, return SI-Units
        result_variable = Variable("{}{}".format(result_value, result_units))

        # If the units of both operands are the same, return those units
        if self.units == other.units:
            result_variable.rescale_to(self.units)

        return result_variable

    @aedt_exception_handler
    def __sub__(self, other):
        """Subtract another variable from the variable to return a new object.

        Parameters
        ---------
        other : Variable
            Object to be subtracted.

        Returns
        -------
        type
            Variable.

        Examples
        --------

        >>> from pyaedt.application.Variables import Variable
        >>> v3 = Variable("3mA")
        >>> v4 = Variable("10A")
        >>> result_2 = v3 - v4
        >>> assert result_2.numeric_value == -9.997
        >>> assert result_2.units == "A"
        >>> assert result_2.unit_system == "Current"

        """
        assert isinstance(other, Variable), "You can only subtract a variable from another variable."
        assert (
            self.unit_system == other.unit_system
        ), "Only ``Variable`` objects with the same unit system can be subtracted."
        result_value = self.value - other.value
        result_units = SI_units[self.unit_system]
        # If the units of the two operands are different, return SI-Units
        result_variable = Variable("{}{}".format(result_value, result_units))

        # If the units of both operands are the same, return those units
        if self.units == other.units:
            result_variable.rescale_to(self.units)

        return result_variable

    # Python 3.x version
    @aedt_exception_handler
    def __truediv__(self, other):
        """Divide the variable by a number or another variable to return a new object.

        Parameters
        ---------
        other : numbers.Number or variable
            Object by which to divide.

        Returns
        -------
        type
            Variable.

        Examples
        --------
        Divide a variable with units ``"W"`` by a variable with units ``"V"`` and automatically
        resolve the new units to ``"A"``.

        >>> from pyaedt.application.Variables import Variable

        >>> v1 = Variable("10W")
        >>> v2 = Variable("40V")
        >>> result = v1 / v2
        >>> assert result_1.numeric_value == 0.25
        >>> assert result_1.units == "A"
        >>> assert result_1.unit_system == "Current"

        """
        assert is_number(other) or isinstance(other, Variable), "Divisor must be a scalar quantity or a variable."
        if is_number(other):
            result_value = self.numeric_value / other
            result_units = self.units
        else:
            result_value = self.value / other.value
            result_units = _resolve_unit_system(self.unit_system, other.unit_system, "divide")

        return Variable("{}{}".format(result_value, result_units))

    # Python 2.7 version
    @aedt_exception_handler
    def __div__(self, other):
        return self.__truediv__(other)

    @aedt_exception_handler
    def __rtruediv__(self, other):
        """Divide another object by this object.

        Parameters
        ---------
        other : numbers.Number or variable
            Object to divide by.

        Returns
        -------
        type
            Variable.

        Examples
        --------
        Divide a number by a variable with units ``"s"`` and automatically determine that
        the result is in ``"Hz"``.

        >>> from pyaedt.application.Core.Variables import Variable
        >>> v = Variable("1s")
        >>> result = 3.0 / v
        >>> assert result.numeric_value == 3.0
        >>> assert result.units == "Hz"
        >>> assert result.unit_system == "Freq"

        """
        assert is_number(other), "Dividend must be a numerical quantity!"
        result_value = other / self.value
        result_units = _resolve_unit_system("None", self.unit_system, "divide")
        return Variable("{}{}".format(result_value, result_units))

    # Python 2.7 version
    @aedt_exception_handler
    def __div__(self, other):
        return self.__rtruediv__(other)


class Expression(Variable, object):
    """Provides a framework for manipulating variable expressions.

    Parameters
    ----------
    expression :

    float_value :

    """

    def __init__(self, expression, float_value, full_variables={}):
        self._expression = expression
        self._value = float_value
        try:
            value, units = decompose_variable_value(expression, full_variables)
            self._units = units
        except:
            self._units = ""

    @property
    def expression(self):
        """Expression."""
        return str(self._expression)


class DataSet(object):
    """Manages datasets.

    Parameters
    ----------
    parent :
    name :
    x : float
    y : float
    z : float, optional
        The default is ``None``.
    v : float, optional
       The default is ``None``.
    xunit : str, optional
       The default is ``""``.
    yunit : str, optional
       The default is ``""``.
    zunit : str, optional
       The default is ``""``.
    vunit : str, optional
       The default is ``""``.

    """

    def __init__(self, parent, name, x, y, z=None, v=None, xunit="", yunit="", zunit="", vunit=""):
        self._parent = parent
        self.name = name
        self.x = x
        self.y = y
        self.z = z
        self.v = v
        self.xunit = xunit
        self.yunit = yunit
        self.zunit = zunit
        self.vunit = vunit

    @aedt_exception_handler
    def _args(self):
        """Retrieve arguments."""
        arg = []
        arg.append("Name:" + self.name)
        arg2 = ["Name:Coordinates"]
        if self.z is None:
            arg2.append(["NAME:DimUnits", self.xunit, self.yunit])
        elif self.v is not None and self.name[0] == "$":
            arg2.append(["NAME:DimUnits", self.xunit, self.yunit, self.zunit, self.vunit])
        else:
            return False
        if self.z and self.name[0] == "$":
            x, y, z, v = (list(t) for t in zip(*sorted(zip(self.x, self.y, self.z, self.v))))
        else:
            x, y = (list(t) for t in zip(*sorted(zip(self.x, self.y))))
        for i in range(len(x)):
            arg3 = []
            arg3.append("NAME:Coordinate")
            arg4 = ["NAME:CoordPoint"]
            arg4.append(float(x[i]))
            arg4.append(float(y[i]))
            if self.z and self.name[0] == "$":
                arg4.append(float(z[i]))
                arg4.append(float(v[i]))
            arg3.append(arg4)
            arg2.append(arg3)
        arg.append(arg2)
        return arg

    @aedt_exception_handler
    def create(self):
        """Create a dataset.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if self.name[0] == "$":
            self._parent._oproject.AddDataset(self._args())
        else:
            self._parent.odesign.AddDataset(self._args())
        return True

    @aedt_exception_handler
    def add_point(self, x, y, z=None, v=None):
        """Add a point to the dataset.

        Parameters
        ----------
        x : float
        y : float
        z : float, optional
            The default is ``None``.
        v : float, optional
            The default is ``None``.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        self.x.append(x)
        self.y.append(y)
        if self.z and self.v:
            self.z.append(z)
            self.v.append(v)

        return self.update()

    @aedt_exception_handler
    def remove_point_from_x(self, x):
        """Remove a point from an X-axis value.

        Parameters
        ----------
        x : float

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if x not in self.x:
            self._parent._messenger.add_error_message("Value {} is not found.".format(x))
            return False
        id_to_remove = self.x.index(x)
        return self.remove_point_from_index(id_to_remove)

    @aedt_exception_handler
    def remove_point_from_index(self, id_to_remove):
        """Remove a point from an index.

        Parameters
        ----------
        id_to_remove : int
            ID of the index.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if id_to_remove < len(self.x) > 2:
            self.x.pop(id_to_remove)
            self.y.pop(id_to_remove)
            if self.z and self.v:
                self.z.pop(id_to_remove)
                self.v.pop(id_to_remove)
            return self.update()
        self._parent._messenger.add_error_message("cannot Remove {} index.".format(id_to_remove))
        return False

    @aedt_exception_handler
    def update(self):
        """Update the dataset.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        args = self._args()
        if not args:
            return False
        if self.name[0] == "$":
            self._parent._oproject.EditDataset(self.name, self._args())
        else:
            self._parent.odesign.EditDataset(self.name, self._args())
        return True

    @aedt_exception_handler
    def delete(self):
        """Delete the dataset.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if self.name[0] == "$":
            self._parent._oproject.DeleteDataset(self.name)
            del self._parent.project_datasets[self.name]
        else:
            self._parent.odesign.DeleteDataset(self.name)
            del self._parent.project_datasets[self.name]
        return True

    @aedt_exception_handler
    def export(self, dataset_path=None):
        """Export the dataset.

        Parameters
        ----------
        dataset_path : str, optional
            Path to export the dataset to. The default is ``None``, in which
            case the dataset is exported to the project path.

        Returns
        -------
        bool
            ``True`` when successful, ``False`` when failed.

        """
        if not dataset_path:
            dataset_path = os.path.join(self._parent.project_path, self.name + ".tab")
        if self.name[0] == "$":
            self._parent._oproject.ExportDataset(self.name, dataset_path)
        else:
            self._parent.odesign.ExportDataset(self.name, dataset_path)
        return True
